# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The module for the wordgame solver."""

from __future__ import annotations

from collections.abc import Collection

from hangman_solver import Language, read_words_with_length
from typed_stream import Stream

from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo, bounded_edit_distance


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            ("/wortspiel-helfer", WordgameSolver),
            ("/api/wortspiel-helfer", WordgameSolverAPI),
        ),
        name="Wortspiel-Helfer",
        description=(
            "Findet Worte, die nur eine Änderung voneinander entfernt sind."
        ),
        path="/wortspiel-helfer",
        keywords=("Wortspiel", "Helfer", "Hilfe", "Worte"),
        aliases=("/wordgame-solver",),
        hidden=True,
    )


def find_solutions(word: str, ignore: Collection[str]) -> Stream[str]:
    """Find words that have only one different letter."""
    word_len = len(word)
    ignore = {*ignore, word}

    return (
        Stream((word_len - 1, word_len, word_len + 1))
        .flat_map(
            lambda length: read_words_with_length(
                Language.DeBasicUmlauts, length
            )
        )
        .exclude(ignore.__contains__)
        .filter(
            lambda test_word: bounded_edit_distance(word, test_word, 2) == 1
        )
    )


def get_ranked_solutions(
    word: str, before: Collection[str] = ()
) -> list[tuple[int, str]]:
    """Find solutions for the word and rank them."""
    if not word:
        return []
    before_with_word = {*before, word}
    return sorted(
        (
            (find_solutions(sol, before_with_word).count(), sol)
            for sol in find_solutions(word, before)
        ),
        reverse=True,
    )


class WordgameSolver(HTMLRequestHandler):
    """The request handler for the wordgame solver page."""

    RATELIMIT_GET_LIMIT = 10

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the wordgame solver page."""
        if head:
            return
        word = self.get_argument("word", "").lower()
        before_str = self.get_argument("before", "")
        before = [_w.strip() for _w in before_str.split(",") if _w.strip()]
        new_before = [*before, word] if word and word not in before else before

        await self.render(
            "pages/wordgame_solver.html",
            word=word,
            words=get_ranked_solutions(word, before),
            before=", ".join(before),
            new_before=", ".join(new_before),
        )


class WordgameSolverAPI(APIRequestHandler):
    """The request handler for the wordgame solver API."""

    RATELIMIT_GET_LIMIT = 10
    ALLOWED_METHODS = ("GET",)

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the wordgame solver API."""
        if head:
            return
        word: str = self.get_argument("word", "").lower()
        before_str: str = self.get_argument("before", "")
        before = [_w.strip() for _w in before_str.split(",") if _w.strip()]
        return await self.finish_dict(
            before=before,
            word=word,
            solutions=get_ranked_solutions(word, before),
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A module with useful decorators."""

from __future__ import annotations

import contextlib
import logging
from base64 import b64decode
from collections.abc import Callable, Mapping
from functools import wraps
from typing import Any, ParamSpec, TypeVar, cast, overload

from tornado.web import RequestHandler

from .token import InvalidTokenError, parse_token
from .utils import Permission, anonymize_ip

Default = TypeVar("Default")
Args = ParamSpec("Args")
Ret = TypeVar("Ret")


def keydecode(
    token: str,
    api_secrets: Mapping[str | None, Permission],
    token_secret: str | bytes | None,
) -> None | Permission:
    """Decode a key."""
    tokens: list[str] = [token]
    decoded: str | None
    try:
        decoded = b64decode(token).decode("UTF-8")
    except ValueError:
        decoded = None
    else:
        tokens.append(decoded)
    if token_secret:
        for _ in tokens:
            with contextlib.suppress(InvalidTokenError):
                return parse_token(_, secret=token_secret).permissions
    if decoded is None:
        return None
    return api_secrets.get(decoded)


def is_authorized(
    inst: RequestHandler,
    permission: Permission,
    allow_cookie_auth: bool = True,
) -> None | bool:
    """Check whether the request is authorized."""
    keys: dict[str | None, Permission] = inst.settings.get(
        "TRUSTED_API_SECRETS", {}
    )
    token_secret: str | bytes | None = inst.settings.get("AUTH_TOKEN_SECRET")

    permissions: tuple[None | Permission, ...] = (  # TODO: CLEAN-UP THIS MESS!!
        *(
            (
                keydecode(_[7:], keys, token_secret)
                if _.lower().startswith("bearer ")
                else keys.get(_)
            )
            for _ in inst.request.headers.get_list("Authorization")
        ),
        *(
            keydecode(_, keys, token_secret)
            for _ in inst.get_arguments("access_token")
        ),
        *(keys.get(_) for _ in inst.get_arguments("key")),
        (
            keydecode(
                inst.get_cookie("access_token", ""),
                keys,
                token_secret,
            )
            if allow_cookie_auth
            else None
        ),
        keys.get(inst.get_cookie("key", None)) if allow_cookie_auth else None,
    )

    if all(perm is None for perm in permissions):
        return None

    result = Permission(0)
    for perm in permissions:
        if perm:
            result |= perm

    return permission in result


_DefaultValue = object()


@overload
def requires(
    *perms: Permission,
    return_instead_of_finishing: Default,
    allow_cookie_auth: bool = True,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Ret | Default]]: ...


@overload
def requires(
    *perms: Permission,
    allow_cookie_auth: bool = True,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Ret]]: ...


def requires(
    *perms: Permission,
    return_instead_of_finishing: Any = _DefaultValue,
    allow_cookie_auth: bool = True,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Any]]:
    """Handle required permissions."""
    permissions = Permission(0)
    for perm in perms:
        permissions |= perm

    finish_with_error = return_instead_of_finishing is _DefaultValue

    def internal(method: Callable[Args, Ret]) -> Callable[Args, Any]:
        method.required_perms = permissions  # type: ignore[attr-defined]
        logger = logging.getLogger(f"{method.__module__}.{method.__qualname__}")

        wraps(method)

        @wraps(method)
        def wrapper(*args: Args.args, **kwargs: Args.kwargs) -> Any:
            instance = args[0]
            if not isinstance(instance, RequestHandler):
                raise TypeError(f"Instance has invalid type {type(instance)}")
            authorized = is_authorized(instance, permissions, allow_cookie_auth)
            if not authorized:
                if not finish_with_error:
                    return cast(Ret, return_instead_of_finishing)
                logger.warning(
                    "Unauthorized access to %s from %s",
                    instance.request.path,
                    anonymize_ip(instance.request.remote_ip),
                )
                instance.clear()
                instance.set_header("WWW-Authenticate", "Bearer")
                status = 401 if authorized is None else 403
                instance.set_status(status)
                instance.write_error(status, **kwargs)
                return None

            return method(*args, **kwargs)

        return wrapper

    return internal


@overload
def requires_settings(
    *settings: str,
    return_: Default,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Ret | Default]]: ...


@overload
def requires_settings(
    *settings: str,
    status_code: int,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Ret | None]]: ...


def requires_settings(
    *settings: str,
    return_: Any = _DefaultValue,
    status_code: int | None = None,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Any]]:
    """Require some settings to execute a method."""
    finish_with_error = return_ is _DefaultValue
    if not finish_with_error and isinstance(status_code, int):
        raise ValueError("return_ and finish_status specified")
    if finish_with_error and status_code is None:
        status_code = 503

    def internal(method: Callable[Args, Ret]) -> Callable[Args, Any]:
        logger = logging.getLogger(f"{method.__module__}.{method.__qualname__}")

        @wraps(method)
        def wrapper(*args: Args.args, **kwargs: Args.kwargs) -> Any:
            instance = args[0]
            if not isinstance(instance, RequestHandler):
                raise TypeError(f"Instance has invalid type {type(instance)}")
            missing = [
                setting
                for setting in settings
                if instance.settings.get(setting) is None
            ]
            if missing:
                if not finish_with_error:
                    return cast(Ret, return_)
                logger.warning(
                    "Missing settings %s for request to %s",
                    ", ".join(missing),
                    instance.request.path,
                )
                instance.send_error(cast(int, status_code))
                return None
            for setting in settings:
                kwargs[setting.lower()] = instance.settings[setting]
            return method(*args, **kwargs)

        return wrapper

    return internal


def get_setting_or_default(
    setting: str,
    default: Any,
) -> Callable[[Callable[Args, Ret]], Callable[Args, Ret]]:
    """Require some settings to execute a method."""

    def internal(method: Callable[Args, Ret]) -> Callable[Args, Ret]:
        @wraps(method)
        def wrapper(*args: Args.args, **kwargs: Args.kwargs) -> Ret:
            instance = args[0]
            if not isinstance(instance, RequestHandler):
                raise TypeError(f"Instance has invalid type {type(instance)}")
            kwargs[setting.lower()] = instance.settings.get(setting, default)
            return method(*args, **kwargs)

        return wrapper

    return internal



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Get cool commit messages.

Based on: https://github.com/ngerakines/commitment
"""

from __future__ import annotations

import logging
import random
from collections.abc import Mapping
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Final

import emoji
from tornado.web import HTTPError
from typed_stream import Stream

from .. import DIR as ROOT_DIR
from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)

type Commit = tuple[datetime, str]
type Commits = Mapping[str, Commit]


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/commitment", CommitmentAPI),),
        name="Commitment",
        short_name="Commitment",
        description="Zeige gute Commit-Nachrichten an.",
        path="/api/commitment",
        aliases=(),
        sub_pages=(),
        keywords=(),
        hidden=True,
    )


def parse_commits_txt(data: str) -> Commits:
    """Parse the contents of commits.txt."""
    return {
        split[0]: (datetime.fromtimestamp(int(split[1]), UTC), split[2])
        for line in data.splitlines()
        if (split := line.rstrip().split(" ", 2))
    }


def read_commits_txt() -> None | Commits:
    """Read the contents of the local commits.txt file."""
    if not (file := ROOT_DIR / "static" / "commits.txt").is_file():
        return None
    return parse_commits_txt(file.read_text("UTF-8"))


COMMITS: None | Commits = read_commits_txt()


@dataclass(slots=True)
class Arguments:
    """The arguments for the commitment API."""

    hash: str | None = None
    require_emoji: bool = False


class CommitmentAPI(APIRequestHandler):
    """The request handler for the commitment API."""

    POSSIBLE_CONTENT_TYPES = (
        "text/plain",
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    @parse_args(type_=Arguments)
    async def get(self, *, args: Arguments, head: bool = False) -> None:
        """Handle GET requests to the API."""
        # pylint: disable=unused-argument
        if not COMMITS:
            raise HTTPError(
                503,
                log_message="No COMMITS found, make sure to create commits.txt",
            )

        if args.hash is None:
            return await self.write_commit(
                *random.choice(
                    [
                        (com, (_, msg))
                        for com, (_, msg) in COMMITS.items()
                        if not args.require_emoji or any(emoji.analyze(msg))
                    ]
                )
            )

        if len(args.hash) + 2 == 42:
            if args.hash in COMMITS:
                return await self.write_commit(args.hash, COMMITS[args.hash])
            raise HTTPError(404)

        if len(args.hash) + 1 >= 42:
            raise HTTPError(404)

        results = (
            Stream(
                (com, (_, msg))
                for com, (_, msg) in COMMITS.items()
                if com.startswith(args.hash)
                if not args.require_emoji or any(emoji.analyze(msg))
            )
            .limit(2)
            .collect()
        )

        if len(results) != 1:
            raise HTTPError(404)

        [(hash_, commit)] = results

        return await self.write_commit(hash_, commit)

    async def write_commit(self, hash_: str, commit: Commit) -> None:
        """Write the commit data."""
        self.set_header("X-Commit-Hash", hash_)

        if self.content_type == "text/plain":
            return await self.finish(commit[1])

        return await self.finish_dict(
            hash=hash_,
            commit_message=commit[1],
            permalink=self.fix_url("/api/commitment", hash=hash_),
            date=commit[0],
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that converts german currencies."""

from __future__ import annotations

import random
from typing import Final, TypeAlias, cast

import regex

from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo

ValueDict: TypeAlias = dict[str, str | int | bool]
ValuesTuple: TypeAlias = tuple[int, int, int, int]

KEYS: Final = ("euro", "mark", "ost", "schwarz")
MULTIPLIERS: Final = (1, 2, 4, 20)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/waehrungs-rechner", CurrencyConverter),
            (r"/api/waehrungs-rechner", CurrencyConverterAPI),
        ),
        name="Währungsrechner",
        description=(
            "Ein Währungsrechner für teilweise veraltete deutsche Währungen"
        ),
        path="/waehrungs-rechner",
        keywords=(
            "Währungsrechner",
            "converter",
            "Euro",
            "D-Mark",
            "Ost-Mark",
            "Känguru",
        ),
        aliases=(
            "/rechner",
            "/w%C3%A4hrungs-rechner",
            "/währungs-rechner",
            "/currency-converter",
        ),
    )


def string_to_num(num: str, divide_by: int = 1) -> int:
    """Convert a string to a number and divide it by divide_by."""
    num = regex.sub(r"[^\d,]", "", num.replace(".", ","))

    if not num:
        return 0

    if "," not in num:
        return (int(num) * 100) // divide_by

    split = [spam for spam in num.split(",") if spam]
    if not split:
        return 0

    if len(split) == 1:
        return (int(split[0]) * 100) // divide_by

    dec = split[-1]
    pre = "".join(split[:-1])

    return (int(pre or "0") * 100 + int((dec + "00")[0:2])) // divide_by


def num_to_string(num: int) -> str:
    """
    Convert a float to the german representation of a number.

    The number has 2 or 0 digits after the comma.
    """
    if not num:
        return "0"
    string = f"00{num}"
    string = (string[0:-2].lstrip("0") or "0") + "," + string[-2:]
    return string.removesuffix(",00")


async def conversion_string(value_dict: ValueDict) -> str:
    """Generate a text that complains how expensive everything is."""
    return (
        f"{value_dict.get('euro_str')} Euro, "
        f"das sind ja {value_dict.get('mark_str')} Mark; "
        f"{value_dict.get('ost_str')} Ostmark und "
        f"{value_dict.get('schwarz_str')} Ostmark auf dem Schwarzmarkt!"
    )


async def continuation_string(
    values: ValuesTuple,
    ticket_price: None | int = None,
    ins_kino_gehen: None | str = None,
) -> str:
    """Generate a second text that complains how expensive everything is."""
    price: int = ticket_price or values[0]
    if not ins_kino_gehen:
        ins_kino_gehen = "ins Kino gehen"
    price_ostmark: int = (
        100
        if ins_kino_gehen == "ins Kino gehen"
        else int(max(round(price // 8) // 2, 50))
    )
    rand = random.Random(f"{price}|{values}|{ins_kino_gehen}")
    kino_count: int = values[-1] // price_ostmark
    output = [
        "Und ich weiß nicht, ob ihr das noch wisst, aber man konnte locker für",
        "eine" if price_ostmark == 100 else num_to_string(price_ostmark),
        (
            f"Ostmark {ins_kino_gehen}! Das heißt man konnte von "
            f"{num_to_string(values[-1])} Ostmark "
            f"{kino_count}-mal {ins_kino_gehen}."
        ),
    ]
    while True:  # pylint: disable=while-used
        euro, mark, ost, schwarz = convert(int(price * kino_count))
        no_time = (
            " — dafür habt ihr ja keine Zeit im Kapitalismus, "
            "aber wenn ihr die Zeit hättet —"
            if mark > 20_300_000_000 * 100
            else ""
        )
        output.append(
            f"Wenn ihr aber heute {kino_count}-mal "
            f"{ins_kino_gehen} wollt{no_time}, müsst ihr"
        )
        if euro > 1_000_000 * 100:
            output.append("...äh...")
        output.append(f"{num_to_string(euro)} Euro bezahlen.")
        output.append(
            rand.choice(
                (
                    "Ja! Ja! Ich weiß, was ihr denkt!",
                    "Ja, ja, ich weiß, was ihr denkt!",
                    "Ich weiß, was ihr denkt!",
                )
            )
        )
        if mark > 20_300_000_000 * 100:  # Staatsschulden der DDR
            output.append(  # the end of the text
                "Davon hätte man die DDR entschulden können! "
                f"Von einmal {ins_kino_gehen}. "
                "So teuer ist das alles geworden."
            )
            break
        new_kino_count = int((schwarz * price_ostmark) // 100)
        # TODO: add random chance to get approximation
        output.append(
            f"{num_to_string(mark)} Mark, {num_to_string(ost)} Ostmark, "
            f"{num_to_string(schwarz)} Ostmark auf dem Schwarzmarkt, "
            f"davon konnte man {new_kino_count}-mal {ins_kino_gehen}."
        )
        if new_kino_count <= kino_count:
            # it doesn't grow, exit to avoid infinite loop
            break
        kino_count = new_kino_count

    return " ".join(output)


def convert(euro: int) -> ValuesTuple:
    """Convert a number to the german representation of a number."""
    return cast(ValuesTuple, tuple(euro * _m for _m in MULTIPLIERS))


async def get_value_dict(
    euro: int,
    ins_kino_gehen: None | str = None,
) -> ValueDict:
    """Create the value dict base on the euro."""
    values = convert(euro)
    value_dict: ValueDict = {}
    for key, val in zip(KEYS, values, strict=True):
        value_dict[key] = val
        value_dict[key + "_str"] = num_to_string(val)

    value_dict["text"] = await conversion_string(value_dict)
    value_dict["text2"] = await continuation_string(
        values, ins_kino_gehen=ins_kino_gehen
    )
    return value_dict


class CurrencyConverter(HTMLRequestHandler):
    """Request handler for the currency converter page."""

    async def create_value_dict(self) -> None | ValueDict:
        """
        Parse the arguments to get the value dict and return it.

        Return None if no argument is given.
        """
        arg_list: list[tuple[int, str, str]] = []

        for i, key in enumerate(KEYS):
            num_str = self.get_argument(key, None)
            if num_str:
                arg_list.append((i, key, num_str))

        too_many_params = len(arg_list) > 1

        for i, key, num_str in arg_list:
            euro = string_to_num(num_str, MULTIPLIERS[i])
            if euro is not None:
                value_dict = await get_value_dict(
                    euro, ins_kino_gehen=self.get_argument("text", None)
                )
                if too_many_params:
                    value_dict["too_many_params"] = True
                value_dict["key_used"] = key
                return value_dict

        return None

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the currency converter page."""
        # pylint: disable=unused-argument
        value_dict = await self.create_value_dict()
        if value_dict is None:
            value_dict = await get_value_dict(0)
            description = self.description
        else:
            description = cast(str, value_dict["text"])

        if value_dict.get("too_many_params", False):
            used_key = value_dict.get("key_used")
            replace_url_with = self.fix_url(
                self.request.full_url(),
                **{  # type: ignore[arg-type]
                    key: (
                        value_dict.get(f"{used_key}_str")
                        if key == used_key
                        else None
                    )
                    for key in KEYS
                },
            )
            if replace_url_with != self.request.full_url():
                return self.redirect(replace_url_with)

        await self.render(
            "pages/converter.html",
            **value_dict,
            description=description,
        )


class CurrencyConverterAPI(APIRequestHandler, CurrencyConverter):
    """Request handler for the currency converter API."""

    async def get(self, *, head: bool = False) -> None:
        """
        Handle GET requests to the currency converter API.

        If no arguments are given the potential arguments are shown.
        """
        value_dict = await self.create_value_dict()

        if head:
            return

        if value_dict is None:
            return await self.finish(
                dict(zip(KEYS, [None] * len(KEYS), strict=True))
            )

        for key in KEYS:
            value_dict[key] = value_dict.pop(f"{key}_str")

        return await self.finish(value_dict)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the swapped words page."""

from __future__ import annotations

from base64 import b64encode
from datetime import datetime

import orjson as json
import pytest
import regex
import tornado.web
import yaml
from time_machine import travel

from an_website.swapped_words import config_file as sw_config, swap

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_html_response,
    assert_valid_json_response,
    fetch,
)
from .test_settings import parse_cookie


def test_copying_case_of_letters() -> None:
    """Test the copying of cases for single letters."""
    lower_char = "a"
    upper_char = "A"
    for _c1, _c2 in (
        (lower_char, lower_char),
        (lower_char, upper_char),
        (upper_char, upper_char),
        (upper_char, lower_char),
    ):
        assert sw_config.copy_case_letter(_c1, _c2) == _c1

    for char in ("a", "B", "c", "D", "e", "F", "g", "H"):
        assert sw_config.copy_case_letter(char, "a").isupper() == char.isupper()
        assert sw_config.copy_case_letter(char, char.lower()) == char


def test_copying_case() -> None:
    """Test the copying of cases for strings."""
    for word in ("lower", "Upper", "CAPS", "AlTeRnAtInG", "lasT", "AHh!1!1!!"):
        assert sw_config.copy_case(word, word.lower()) == word
        assert sw_config.copy_case(word, word.upper()) == word
        assert sw_config.copy_case(word, word.title()) == word

    for string in (
        "abc def",
        "abc-def",
        "a b c d e f",
        "¯\\_(ツ)_/¯",
        "ab, cd.",
    ):
        lower = string.lower()
        upper = string.upper()
        title = string.title()
        for string in (  # pylint: disable=redefined-loop-name
            lower,
            upper,
            title,
        ):
            assert sw_config.copy_case(string, lower) == string
            assert sw_config.copy_case(string, upper) == string
            assert sw_config.copy_case(string, title) == string

    assert sw_config.copy_case("Long Short", "short LONG") == "Short Long"
    assert sw_config.copy_case("Test1 TEST2", "word1 word2") == "Word1 WORD2"


def test_parsing_config() -> None:
    """Test parsing the sw-config."""
    str_to_replace = "A a B b  XX  Cc cc cC CC  XX  Dd dd dD DD"

    test_conf_str = "(a)=>b;Cc<=>Dd"
    parsed_conf = sw_config.SwappedWordsConfig(test_conf_str)

    assert (  # pylint: disable=unnecessary-dunder-call
        parsed_conf.__eq__(None) == NotImplemented
    )
    beautified = parsed_conf.to_config_str()
    minified = parsed_conf.to_config_str(minified=True)

    assert minified == test_conf_str
    assert parsed_conf == sw_config.SwappedWordsConfig(beautified)
    assert (
        parsed_conf.swap_words(str_to_replace)
        == "B b B b  XX  Dd dd dD DD  XX  Cc cc cC CC"
    )
    assert regex.fullmatch(parsed_conf.get_regex(), "a")
    assert regex.fullmatch(parsed_conf.get_regex(), "cc")
    assert sw_config.beautify(minified) == beautified
    assert sw_config.minify(beautified) == minified
    assert parsed_conf.get_replacement_by_group_name("n", "lol") == "lol"
    assert parsed_conf.get_replacement_by_group_name("lol", "lol") == "lol"

    test_conf_str = "a  <=> b\nCc  => Dd"
    parsed_conf = sw_config.SwappedWordsConfig(test_conf_str)

    beautified = parsed_conf.to_config_str()
    minified = parsed_conf.to_config_str(minified=True)

    assert beautified == test_conf_str
    assert parsed_conf == sw_config.SwappedWordsConfig(minified)
    assert (
        parsed_conf.swap_words(str_to_replace)
        == "B b A a  XX  Dd dd dD DD  XX  Dd dd dD DD"
    )
    assert regex.fullmatch(parsed_conf.get_regex(), "a")
    assert regex.fullmatch(parsed_conf.get_regex(), "cc")
    assert sw_config.beautify(minified) == beautified
    assert sw_config.minify(beautified) == minified

    assert not sw_config.SwappedWordsConfig("").to_config_str()

    # test invalid configs
    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig(" <=> b")

    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig("a <=> ")

    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig("a <> b")

    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig("a <= b")

    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig("a( => b")

    with pytest.raises(sw_config.InvalidConfigError):
        sw_config.SwappedWordsConfig("a <=> (b")

    try:
        sw_config.SwappedWordsConfig("a <=> (b")
    except sw_config.InvalidConfigError as exc:
        assert exc.reason in str(exc)


def test_two_way_word_pair() -> None:
    """Test the two-way word pair."""
    # pylint: disable=too-many-function-args
    pair = sw_config.TwoWayPair("x", "y")  # x <=> y
    assert pair.to_pattern_str() == "x|y"
    assert pair.get_replacement("x") == "y"
    assert pair.get_replacement("X") == "Y"
    assert pair.get_replacement("y") == "x"
    assert pair.get_replacement("Y") == "X"
    assert pair.get_replacement("z") == "z"
    assert pair.get_replacement("Z") == "Z"


def test_one_way_word_pair() -> None:
    """Test the one-way word pair."""
    # pylint: disable=too-many-function-args
    pair = sw_config.OneWayPair("x", "y")  # x => y
    assert pair.to_pattern_str() == "x"
    assert pair.get_replacement("x") == "y"
    assert pair.get_replacement("X") == "Y"
    assert pair.get_replacement("y") == "y"
    assert pair.get_replacement("Y") == "Y"
    assert pair.get_replacement("z") == "z"
    assert pair.get_replacement("Z") == "Z"


def test_check_text_too_long() -> None:
    """Test the check_text_too_long function."""
    swap.check_text_too_long("")
    swap.check_text_too_long("x" * swap.MAX_CHAR_COUNT)
    with pytest.raises(tornado.web.HTTPError):
        swap.check_text_too_long("x" * (swap.MAX_CHAR_COUNT + 1))


@travel(datetime(1, 2, 3, 4, 5, 6, 7), tick=False)
async def test_sw_html_request_handlers(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test the swapped words html request handlers."""
    assert_valid_html_response(
        await fetch(
            "/vertauschte-woerter",
            method="POST",
            body="reset=nope&text=" + "x" * (swap.MAX_CHAR_COUNT + 1),
        ),
        {413},  # text too long
    )
    assert_valid_html_response(
        await fetch(
            "/vertauschte-woerter",
            method="POST",
            body="",
        ),
        {400},  # missing argument
    )
    assert_valid_html_response(
        await fetch(
            "/vertauschte-woerter",
            method="POST",
            body="reset=nope&text=x&config=xyz",
        ),
        {400},  # invalid config
    )
    # everything ok
    assert_valid_html_response(
        await fetch(
            "/vertauschte-woerter",
            method="POST",
            body="reset=nope&text=x",
        ),
    )

    response = assert_valid_html_response(
        await fetch(
            "/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/yaml"},
            body=yaml.dump(
                {"reset": "nope", "text": "text", "config": "abc <=> xyz"}
            ),
        )
    )
    cookies = response.headers.get_list("Set-Cookie")
    assert len(cookies) == 1
    cookie = parse_cookie(cookies[0])
    morsel = cookie[tuple(cookie)[0]]
    assert morsel["expires"]
    assert morsel["path"] == "/vertauschte-woerter"
    assert morsel["samesite"] == "Strict"
    assert morsel.key == "swapped-words-config"
    assert morsel.value == b64encode(b"abc <=> xyz").decode("UTF-8")

    response2 = await fetch(
        "/vertauschte-woerter",
        method="POST",
        headers={"Content-Type": "application/json", "Cookie": cookies[0]},
        body=json.dumps(
            {
                "reset": "nope",
                "text": "text",
            }
        ),
    )
    assert not response2.headers.get_list("Set-Cookie")
    assert response2.body == response.body


async def test_sw_json_request_handlers(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test the swapped words JSON API request handlers."""
    response = assert_valid_json_response(
        await fetch(
            "/api/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    "text": "x z o",
                    "config": "x  => y\nz <=> o",
                    "return_config": "sure",
                    "minify_config": "sure",
                }
            ),
        ),
    )
    assert response["text"] == "x z o"
    assert response["replaced_text"] == "y o z"
    assert response["return_config"] is True
    assert response["minify_config"] is True
    assert response["config"] == "x=>y;z<=>o"

    response = assert_valid_json_response(
        await fetch(
            "/api/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    "text": "x y z",
                    "config": "# \n",
                    "return_config": "sure",
                    "minify_config": "nope",
                }
            ),
        ),
    )
    assert response["text"] == "x y z"
    assert response["replaced_text"] == "x y z"
    assert response["return_config"] is True
    assert response["minify_config"] is False
    # pylint: disable-next=use-implicit-booleaness-not-comparison-to-string
    assert response["config"] == ""

    response = assert_valid_json_response(
        await fetch(
            "/api/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    "text": "x z o",
                    "config": "x  => y\nz <=> o",
                    "return_config": True,
                    "minify_config": False,
                }
            ),
        ),
    )
    assert response["text"] == "x z o"
    assert response["replaced_text"] == "y o z"
    assert response["return_config"] is True
    assert response["minify_config"] is False
    assert response["config"] == "x  => y\nz <=> o"

    response = assert_valid_json_response(
        await fetch(
            "/api/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    "config": "x  => y\nz <> o",
                }
            ),
        ),
        {400},
    )
    assert response["line"] == "z <> o"
    assert response["line_num"] == 1

    response = assert_valid_json_response(
        await fetch(
            "/api/vertauschte-woerter",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    "config": "x  => y\nz <=> o\na == b",
                }
            ),
        ),
        {400},
    )
    assert response["line"] == "a == b"
    assert response["line_num"] == 2


if __name__ == "__main__":
    test_copying_case_of_letters()
    test_copying_case()
    test_parsing_config()
    test_two_way_word_pair()
    test_one_way_word_pair()
    test_check_text_too_long()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Stores that contain the ids of old quotes of the day."""

from __future__ import annotations

import abc
from datetime import date, datetime, timezone
from typing import ClassVar, Final

from redis.asyncio import Redis
from tornado.web import HTTPError

from ... import EVENT_REDIS

QUOTE_COUNT_TO_SHOW_IN_FEED: Final[int] = 8


class QuoteOfTheDayStore(abc.ABC):
    """The class representing the store for the quote of the day."""

    __slots__ = ()

    @abc.abstractmethod
    async def get_quote_id_by_date(self, date_: date) -> tuple[int, int] | None:
        """Get the quote id for the given date."""
        raise NotImplementedError

    @abc.abstractmethod
    async def has_quote_been_used(self, quote_id: tuple[int, int]) -> bool:
        """Check if the quote has been used already."""
        raise NotImplementedError

    @abc.abstractmethod
    async def set_quote_id_by_date(
        self, date_: date, quote_id: tuple[int, int]
    ) -> None:
        """Set the quote id for the given date."""
        raise NotImplementedError

    @abc.abstractmethod
    async def set_quote_to_used(self, quote_id: tuple[int, int]) -> None:
        """Set the quote as used."""
        raise NotImplementedError


class QuoteOfTheDayStoreWithCache(QuoteOfTheDayStore, abc.ABC):
    """Quote of the day store with an in memory cache."""

    # pylint: disable=abstract-method

    CACHE: ClassVar[dict[date, tuple[int, int]]]

    @classmethod
    def _get_quote_id_from_cache(cls, date_: date) -> None | tuple[int, int]:
        """Get a quote_id from the cache if it is present."""
        return cls.CACHE.get(date_)

    @classmethod
    def _populate_cache(cls, date_: date, quote_id: tuple[int, int]) -> None:
        """Populate the cache for the quote of today."""
        today = datetime.now(timezone.utc).date()
        # old entries are rarely used, they don't need to be cached
        if (today - date_).days > QUOTE_COUNT_TO_SHOW_IN_FEED:
            cls.CACHE[date_] = quote_id

        for key in tuple(cls.CACHE):
            # remove old entries from cache to save memory
            if (today - key).days > QUOTE_COUNT_TO_SHOW_IN_FEED:
                del cls.CACHE[key]


class RedisQuoteOfTheDayStore(QuoteOfTheDayStoreWithCache):
    """A quote of the day store that stores the quote of the day in Redis."""

    __slots__ = ("redis", "redis_prefix")

    CACHE: ClassVar[dict[date, tuple[int, int]]] = {}

    redis_prefix: str
    redis: Redis[str]

    def __init__(self, redis: Redis[str], redis_prefix: str) -> None:
        """Initialize the Redis quote of the day store."""
        super().__init__()
        self.redis = redis
        self.redis_prefix = redis_prefix

    async def get_quote_id_by_date(self, date_: date) -> tuple[int, int] | None:
        """Get the quote id for the given date."""
        if date_ in self.CACHE:
            return self.CACHE[date_]
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)
        wq_id = await self.redis.get(self.get_redis_quote_date_key(date_))
        if not wq_id:
            return None
        quote, author = wq_id.split("-")
        quote_id = int(quote), int(author)
        self._populate_cache(date_, quote_id)
        return quote_id

    def get_redis_quote_date_key(self, wq_date: date) -> str:
        """Get the Redis key for getting quotes by date."""
        return f"{self.redis_prefix}:quote-of-the-day:by-date:{wq_date.isoformat()}"

    def get_redis_used_key(self, wq_id: tuple[int, int]) -> str:
        """Get the Redis used key."""
        str_id = "-".join(map(str, wq_id))  # pylint: disable=bad-builtin
        return f"{self.redis_prefix}:quote-of-the-day:used:{str_id}"

    async def has_quote_been_used(self, quote_id: tuple[int, int]) -> bool:
        """Check if the quote has been used already."""
        if quote_id in self.CACHE.values():
            return True
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)
        return bool(await self.redis.get(self.get_redis_used_key(quote_id)))

    async def set_quote_id_by_date(
        self, date_: date, quote_id: tuple[int, int]
    ) -> None:
        """Set the quote id for the given date."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)
        await self.redis.setex(
            self.get_redis_quote_date_key(date_),
            60 * 60 * 24 * 420,  # TTL
            "-".join(map(str, quote_id)),  # pylint: disable=bad-builtin
        )
        self._populate_cache(date_, quote_id)

    async def set_quote_to_used(self, quote_id: tuple[int, int]) -> None:
        """Set the quote as used."""
        if not EVENT_REDIS.is_set():
            return

        await self.redis.setex(
            self.get_redis_used_key(quote_id),
            #  we have over 720 funny wrong quotes, so 420 should be ok
            60 * 60 * 24 * 420,  # TTL
            1,  # True
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A permanent redirect to an invite of the Discord guild."""

from __future__ import annotations

import time
from typing import Final

import orjson as json
from tornado.httpclient import AsyncHTTPClient
from tornado.web import HTTPError

from .. import CA_BUNDLE_PATH
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo

GUILD_ID: Final[str] = "367648314184826880"

INVITE_CACHE: dict[
    str,
    tuple[float, str, str] | tuple[float, HTTPError],
] = {}


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/discord", ANDiscord),
            (r"/api/discord", ANDiscordAPI),
            (f"/api/discord/({GUILD_ID})", ANDiscordAPI),
            (r"/api/discord/(\d+)", DiscordAPI),
        ),
        name="Discord-Einladung",
        short_name="Discord",
        description="Eine permanente Einladung zu unserer Discord-Gilde",
        path="/discord",
        keywords=(
            "Discord",
            "Server",
            "Guild",
            "Gilde",
            "Invite",
            "Einladung",
        ),
    )


async def url_returns_200(url: str) -> bool:
    """Check whether a URL returns a status code of 200."""
    response = await AsyncHTTPClient().fetch(
        url,
        method="HEAD",
        raise_error=False,
        ca_certs=CA_BUNDLE_PATH,
    )
    return response.code == 200


async def get_invite(guild_id: str = GUILD_ID) -> tuple[str, str]:
    """
    Get the invite to a Discord guild and return it with the source.

    How to get the invite:
        - from the widget (has to be enabled in guild settings)
        - from DISBOARD (the guild needs to set it up first)

    If the invite couldn't be fetched an HTTPError is raised.
    """
    reason = "Invite not found."

    # try getting the invite from the widget
    url = f"https://discord.com/api/guilds/{guild_id}/widget.json"
    response = await AsyncHTTPClient().fetch(
        url, raise_error=False, ca_certs=CA_BUNDLE_PATH
    )
    if response.code == 200:
        response_json = json.loads(response.body)
        if invite := response_json["instant_invite"]:
            return invite, url
        reason = f"No instant invite in widget ({url}) found."

    # try getting the invite from DISBOARD
    url = f"https://disboard.org/site/get-invite/{guild_id}"
    response = await AsyncHTTPClient().fetch(
        url, raise_error=False, ca_certs=CA_BUNDLE_PATH
    )
    if response.code == 200:
        return (
            json.loads(response.body),
            f"https://disboard.org/server/{guild_id}",
        )

    # check if Top.gg lists the guild
    url = f"https://top.gg/servers/{guild_id}/join"
    if await url_returns_200(url):
        return url, f"https://top.gg/servers/{guild_id}/"

    # check if Discords.com lists the guild
    if await url_returns_200(
        # API endpoint that only returns 200 if the guild exists
        f"https://discords.com/api-v2/server/{guild_id}/relevant"
    ):
        return (
            f"https://discords.com/servers/{guild_id}/join",
            f"https://discords.com/servers/{guild_id}/",
        )

    raise HTTPError(404, reason=reason)


async def get_invite_with_cache(
    guild_id: str = GUILD_ID,
) -> tuple[str, str]:
    """Get an invite from cache or from get_invite()."""
    if guild_id in INVITE_CACHE:
        cache_entry = INVITE_CACHE[guild_id]
        if cache_entry[0] > time.monotonic() - 300:
            if isinstance(cache_entry[1], HTTPError):
                raise cache_entry[1]
            return cache_entry[1], cache_entry[2]

    try:
        invite, source = await get_invite(guild_id)
    except HTTPError as exc:
        INVITE_CACHE[guild_id] = (time.monotonic(), exc)
        raise exc

    INVITE_CACHE[guild_id] = (time.monotonic(), invite, source)

    return invite, source


class ANDiscord(HTMLRequestHandler):
    """The request handler that gets the Discord invite and redirects to it."""

    RATELIMIT_GET_LIMIT = 10

    async def get(self, *, head: bool = False) -> None:
        """Get the Discord invite."""
        invite = (await get_invite_with_cache(GUILD_ID))[0]
        if not self.user_settings.ask_before_leaving:
            return self.redirect(invite)
        if head:
            return
        return await self.render(
            "pages/redirect.html",
            send_referrer=True,
            redirect_url=invite,
            from_url=None,
            discord=True,
        )


class DiscordAPI(APIRequestHandler):
    """The API request handler that gets the Discord invite and returns it."""

    RATELIMIT_GET_LIMIT = 5
    RATELIMIT_GET_COUNT_PER_PERIOD = 10

    async def get(
        self, guild_id: str = GUILD_ID, *, head: bool = False
    ) -> None:
        """Get the Discord invite and render it as JSON."""
        # pylint: disable=unused-argument
        invite, source_url = await get_invite_with_cache(guild_id)
        return await self.finish_dict(invite=invite, source=source_url)


class ANDiscordAPI(DiscordAPI):
    """The API request handler only for the AN Discord guild."""

    RATELIMIT_GET_LIMIT = 10
    RATELIMIT_GET_COUNT_PER_PERIOD = 30



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=wrong-import-order, wrong-import-position

"""Utilities used by the tests of an-website."""

from __future__ import annotations

import sys
from os.path import abspath, dirname

from an_website.utils.utils import str_to_bool

# add parent dir to sys.path
# this makes importing an_website possible
DIR = abspath(dirname(__file__))
PARENT_DIR = abspath(dirname(DIR))
sys.path.append(PARENT_DIR)


import warnings

warnings.filterwarnings("ignore", module="defusedxml")


from an_website import patches

patches.apply()


import asyncio
import socket
from collections.abc import Awaitable, Callable, MutableMapping, Set
from datetime import datetime, timezone
from os import environ  # pylint: disable=ungrouped-imports
from pathlib import Path
from typing import Any, Final, cast
from urllib.parse import parse_qsl, urlsplit
from zoneinfo import ZoneInfo

import orjson as json
import pytest
import regex
import time_machine as tm
import yaml
from blake3 import blake3
from lxml import etree
from lxml.html import document_fromstring
from lxml.html.html5parser import HTMLParser
from openmoji_dist import VERSION as OPENMOJI_VERSION
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPResponse
from tornado.web import Application

# pylint: disable=ungrouped-imports
from an_website import EVENT_ELASTICSEARCH, EVENT_REDIS, NAME, UPTIME, main
from an_website.quotes.utils import parse_wrong_quote
from an_website.utils import elasticsearch_setup
from an_website.utils.base_request_handler import TEXT_CONTENT_TYPES
from an_website.utils.better_config_parser import BetterConfigParser

# Same as in ../scripts/fix_static_url_path.py
ERROR_QUERY: Final[str] = "XXX-COULD-NOT-ADD-HASH-XXX"

WRONG_QUOTE_DATA = {
    # https://zitate.prapsschnalinen.de/api/wrongquotes/1
    "id": 1,
    "author": {
        "id": 2,
        "author": "Kim Jong-il",
    },
    "quote": {
        "id": 1,
        "author": {
            "id": 1,
            "author": "Abraham Lincoln",
        },
        "quote": (
            "Frage nicht, was dein Land für dich tun kann, "
            "frage, was du für dein Land tun kannst."
        ),
    },
    "rating": 4,
    "showed": 216,
    "voted": 129,
}

FetchCallable = Callable[..., Awaitable[HTTPResponse]]

# fmt: off
hill_valley = ZoneInfo("America/Los_Angeles")
destination = environ.get("DELOREAN_DESTINATION", "1985-10-26 01:24:00")
delorean = tm.travel(datetime.fromisoformat(destination).replace(tzinfo=hill_valley))
# fmt: on


def pytest_sessionstart(*_: object) -> None:
    """Travel to the past."""
    assert not tm.escape_hatch.is_travelling()
    delorean.start()
    UPTIME.reset()


def pytest_sessionfinish(*_: object) -> None:
    """Travel back to the future."""
    assert tm.escape_hatch.is_travelling()
    delorean.stop()


@pytest.fixture
def app() -> Application:
    """Create the application."""
    assert NAME.endswith("-test")

    config = BetterConfigParser.from_path(Path(DIR, "config.ini"))

    main.ignore_modules(config)
    app = main.make_app(config)  # pylint: disable=redefined-outer-name

    assert isinstance(app, Application)

    app.settings["debug"] = True

    main.apply_config_to_app(app, config)

    es = elasticsearch_setup.setup_elasticsearch(app)
    redis = main.setup_redis(app)

    loop = asyncio.get_event_loop_policy().get_event_loop()

    if es:
        try:
            loop.run_until_complete(es.transport.perform_request("HEAD", "/"))
        except Exception:  # pylint: disable=broad-except
            EVENT_ELASTICSEARCH.clear()
        else:
            if not EVENT_ELASTICSEARCH.is_set():
                loop.run_until_complete(
                    elasticsearch_setup.setup_elasticsearch_configs(
                        es, app.settings["ELASTICSEARCH_PREFIX"]
                    )
                )
            EVENT_ELASTICSEARCH.set()

    if redis:
        try:
            loop.run_until_complete(redis.ping())
        except Exception:  # pylint: disable=broad-except
            EVENT_REDIS.clear()
        else:
            if not EVENT_REDIS.is_set():
                loop.run_until_complete(
                    redis.setex(
                        (
                            f"{app.settings.get('REDIS_PREFIX')}:quote-of-the-day:"
                            f"by-date:{datetime.now(timezone.utc).date().isoformat()}"
                        ),
                        300,
                        "1-2",
                    )
                )
            EVENT_REDIS.set()

    return app


def assert_url_query(url: str, /, **args: None | str) -> None:
    """Assert properties of a URL."""
    split_url = urlsplit(url or "/")
    query_str = split_url.query
    is_static = split_url.path.startswith(
        ("/static/", "/soundboard/files/")
    ) or split_url.path in {"/favicon.png", "/favicon.jxl", "/humans.txt"}
    query: dict[str, str] = (
        dict(parse_qsl(query_str, True, True)) if query_str else {}
    )
    for key, value in args.items():
        if value is None or (is_static and key != "v"):
            assert key not in query
        elif not split_url.path.startswith("/static/js/utils/"):
            assert key in query or print(url, key, value)
            assert query[key] == value or print(url, key, value)


@pytest.fixture
def fetch(
    http_client: AsyncHTTPClient,
    http_server_port: tuple[socket.socket, int],
) -> FetchCallable:
    """Fetch a URL."""
    parse_wrong_quote(WRONG_QUOTE_DATA)
    host = f"http://127.0.0.1:{http_server_port[1]}"

    async def _fetch(
        url: str, *, httpclient: None | AsyncHTTPClient = None, **kwargs: Any
    ) -> HTTPResponse:
        """Fetch a URL."""
        if not url.startswith(("http://", "https://")):
            url = f"{host}/{url.removeprefix('/')}"
        kwargs.setdefault("raise_error", False)
        kwargs.setdefault("follow_redirects", False)
        kwargs.setdefault("use_gzip", False)
        if not kwargs.get("headers"):
            kwargs["headers"] = {}
        kwargs["headers"].setdefault("Accept", "text/html, */*")
        try:
            return await (httpclient or http_client).fetch(url, **kwargs)
        except HTTPClientError:
            print(url, kwargs)
            raise

    return _fetch


async def assert_valid_redirect(
    fetch: FetchCallable,  # pylint: disable=redefined-outer-name
    path: str,
    new_path: str,
    codes: Set[int] = frozenset({307, 308}),
    **kwargs: Any,
) -> HTTPResponse:
    """Assert a valid redirect to a new URL."""
    response = await fetch(path, **kwargs)
    assert response.code in codes or print(path, codes, response.code)
    base_url = response.request.url.removesuffix(path)
    real_new_path = response.headers["Location"].removeprefix(base_url) or "/"
    if real_new_path.startswith("?"):
        real_new_path = f"/{real_new_path}"
    assert real_new_path == new_path or print(path, new_path, real_new_path)
    return response


def assert_valid_response(
    response: HTTPResponse,
    content_type: None | str,
    codes: Set[int] = frozenset({200, 503}),
    headers: None | MutableMapping[str, Any] = None,
    needs_to_end_with_line_feed: bool = True,
) -> HTTPResponse:
    """Assert a valid response with the given status code and Content-Type."""
    url = response.effective_url
    assert response.code in codes or print(
        url, codes, response.code, response.body
    )

    assert ERROR_QUERY.encode("UTF-8") not in response.body or print(url)

    if response.request.method == "HEAD":
        assert not response.body
    else:
        assert int(response.headers["Content-Length"]) == len(response.body)

    if (
        needs_to_end_with_line_feed
        and response.request.method != "HEAD"
        and (
            (
                content_type := cast(
                    str, response.headers.get("Content-Type", "")
                )
            )
            in TEXT_CONTENT_TYPES
            or content_type.startswith("text/")
            or content_type.endswith(("+xml", "+json"))
        )
    ):
        assert response.body.endswith(b"\n") or print(
            f"Body from {url} doesn't end with newline"
        )
    headers = headers or {}
    if content_type:
        headers["Content-Type"] = content_type
    for header, value in headers.items():
        assert response.headers[header] == value or print(
            url, response.headers, header, value
        )
    return response


async def check_html_page(
    fetch: FetchCallable,  # pylint: disable=redefined-outer-name
    url: str | HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
    *,
    recursive: int = 0,
    # pylint: disable=dangerous-default-value
    checked_urls: set[str] = set(),  # noqa: B006
) -> HTTPResponse:
    """Check a HTML page."""
    if isinstance(url, str):
        response = await fetch(url)
    else:
        response = url
        url = response.effective_url
    assert_valid_html_response(response, codes)
    html = document_fromstring(
        response.body.decode("UTF-8"), base_url=response.effective_url
    )
    assert html.find("head") is not None or print("no head found", url)
    assert html.find("body") is not None or print("no body found", url)
    html.make_links_absolute(response.effective_url)
    eff_url = urlsplit(response.effective_url)
    scheme_and_netloc = f"{eff_url.scheme}://{eff_url.netloc}"
    checked_urls.add(url.removeprefix(scheme_and_netloc) or "/")
    found_ref_to_body = False
    responses_to_check: list[HTTPResponse] = []
    for link_tuple in html.iterlinks():
        assert (  # do not allow links to insecure pages
            link_tuple[2].startswith(scheme_and_netloc)
            or link_tuple[2].startswith(
                (
                    "https:",
                    "mailto:",
                    "whatsapp:",
                    "http://📙.la/",
                )
            )
            or urlsplit(link_tuple[2]).netloc.endswith(".onion")
            or print(link_tuple[2], "is not https")
        )
        if (
            # ignore canonical URLs, because they have no query string
            link_tuple[0].tag == "link"
            and link_tuple[1] == "href"
            and link_tuple[0].attrib.get("rel") == "canonical"
            # ignore actions, because the stuff gets set with hidden input
            or link_tuple[1] == "action"
        ):
            continue
        if link_tuple[2] == response.effective_url + "#main":
            found_ref_to_body = True  # every page should have a skip to content
        link: str = link_tuple[2].split("#")[0]
        assert link == link.strip()
        if (
            link.startswith(scheme_and_netloc)
            and (link.removeprefix(scheme_and_netloc) or "/")
            not in checked_urls
        ):
            checked_urls.add(link.removeprefix(scheme_and_netloc) or "/")
            resp = await fetch(link)
            if link.removeprefix(scheme_and_netloc).split("?")[0] in {
                "/discord",
                "/wiki",
            } and not str_to_bool(
                dict(parse_qsl(urlsplit(link).query, True, True)).get(
                    "ask_before_leaving", "0"
                )
            ):
                assert resp.code == 307
                continue
            assert resp.code in {307, *codes, 503}
            resp = assert_valid_response(
                await fetch(link, follow_redirects=True),
                content_type=None,  # ignore Content-Type
                codes=codes,
            )
            if (
                resp.headers["Content-Type"] == "text/html;charset=utf-8"
                and resp.effective_url.startswith(scheme_and_netloc)
                and recursive > 0
            ):
                responses_to_check.append(resp)
            if (
                link.startswith(f"{scheme_and_netloc}/static/")
                or link.startswith(f"{scheme_and_netloc}/soundboard/files/")
                and resp.headers["Content-Type"] != "text/html;charset=utf-8"
            ):
                # check if static file is linked with correct hash
                file_hash = (
                    OPENMOJI_VERSION
                    if link.startswith(f"{scheme_and_netloc}/static/openmoji")
                    else blake3(resp.body).hexdigest(8)
                )
                assert_url_query(link, v=file_hash)
    assert found_ref_to_body or print(url)
    for resp in responses_to_check:
        await check_html_page(
            fetch,
            resp,
            recursive=recursive - 1,
            checked_urls=checked_urls,
            codes=codes,
        )
    return response


def assert_valid_html_response(
    response: HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
    effective_url: None | str = None,
    assert_canonical: bool = True,
) -> HTTPResponse:
    """Assert a valid HTML response with the given status code."""
    assert_valid_response(response, "text/html;charset=utf-8", codes)
    effective_url = effective_url or response.effective_url.split("#")[0]
    body = response.body.decode("UTF-8")
    # check if body is valid HTML5
    spam = HTMLParser(namespaceHTMLElements=False)
    try:
        root = spam.parse(body)
    except Exception:
        print(repr(body))
        raise
    assert not spam.errors or print(effective_url, spam.errors, body)
    if assert_canonical:
        # check if the canonical link is present in the document
        assert (
            (
                url_in_doc := root.find("./head/link[@rel='canonical']").get(
                    "href"
                )
            )
            == effective_url.split("?")[0].rstrip("/")
        ) or print(url_in_doc, effective_url)
    # check for template strings that didn't get replaced
    matches = regex.findall(
        r"{\s*[a-zA-Z_]+\s*}", response.body.decode("UTF-8")
    )
    assert not matches or print(effective_url, matches)
    return response


def assert_valid_rss_response(
    response: HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
) -> etree._Element:
    """Assert a valid RSS response with the given status code."""
    assert_valid_response(response, "application/rss+xml", codes)
    body = response.body
    parsed_xml = etree.fromstring(
        body,
        parser=etree.XMLParser(resolve_entities=False),
        base_url=response.request.url,
    )
    return parsed_xml


def assert_valid_yaml_response(
    response: HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
) -> Any:
    """Assert a valid YAML response with the given status code."""
    assert_valid_response(response, "application/yaml", codes)
    parsed = yaml.full_load(response.body)
    assert parsed
    return parsed


def assert_valid_json_response(
    response: HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
) -> Any:
    """Assert a valid JSON response with the given status code."""
    assert_valid_response(response, "application/json", codes)
    parsed_json = json.loads(response.body)
    assert parsed_json
    return parsed_json


def assert_valid_dynload_response(
    response: HTTPResponse,
    codes: Set[int] = frozenset({200, 503}),
) -> Any:
    """Assert a valid dynload response with the given status code."""
    assert_valid_response(
        response, "application/vnd.asozial.dynload+json", codes
    )
    parsed_json = json.loads(response.body)
    assert parsed_json
    assert response.effective_url == parsed_json["url"]
    assert parsed_json["title"]
    assert "short_title" in parsed_json
    assert parsed_json["body"]
    assert "scripts" in parsed_json
    assert "stylesheets" in parsed_json
    assert "css" in parsed_json
    return parsed_json



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The website of the AN."""

from __future__ import annotations

import multiprocessing
import os
import sys
import time
from asyncio import Event
from collections.abc import Mapping
from importlib.metadata import Distribution
from importlib.resources import files
from importlib.resources.abc import Traversable
from pathlib import Path
from typing import Final, TypedDict

try:
    import orjson as json

    if "spam" not in json.loads('{"spam":"eggs"}'):
        from marshal import dumps, loads

        _loads = json.loads
        json.loads = lambda *args, **kwargs: loads(  # nosec: B302
            dumps(_loads(*args, **kwargs))
        )
except ModuleNotFoundError:
    from . import fake_orjson as json  # type: ignore[no-redef]  # noqa: F811

    sys.modules["orjson"] = json

try:
    from pytest_is_running import is_running as pytest_is_running
except ModuleNotFoundError:

    def pytest_is_running() -> bool:  # noqa: D103
        # pylint: disable=missing-function-docstring
        return "pytest" in sys.modules


class MediaType(TypedDict, total=False):
    # pylint: disable=missing-class-docstring
    charset: str
    compressible: bool
    extensions: list[str]
    source: str


class UptimeTimer:
    """UptimeTimer class used for timing the uptime."""

    __slots__ = ("_start_time",)

    def __init__(self) -> None:
        self.reset()

    def get(self) -> float:
        """Get the time since start in seconds."""
        return self.get_ns() / 1_000_000_000

    def get_ns(self) -> int:
        """Get the time since start in nanoseconds."""
        return time.monotonic_ns() - self._start_time

    def reset(self) -> None:
        """Reset the timer."""
        self._start_time = time.monotonic_ns()


UPTIME: Final = UptimeTimer()

EPOCH: Final[int] = 1651075200
EPOCH_MS: Final[int] = EPOCH * 1000

DIR: Final[Traversable] = files(__name__)

MEDIA_TYPES: Final[Mapping[str, MediaType]] = json.loads(
    (DIR / "vendored" / "mime-db.json").read_bytes()
)

NAME = "an-website"


def get_version() -> str:
    """Get the version of the package."""
    if isinstance(DIR, Path):
        # pylint: disable-next=import-outside-toplevel
        from get_version import get_version as gv

        return gv(__file__, vcs="git")

    return Distribution.from_name(NAME).version


VERSION: Final[str] = get_version()

GH_ORG_URL: Final[str] = "https://github.com/asozialesnetzwerk"
GH_REPO_URL: Final[str] = f"{GH_ORG_URL}/{NAME}"
GH_PAGES_URL: Final[str] = f"https://github.asozial.org/{NAME}"

CACHE_DIR: Final[Path] = Path("~/.cache/", NAME).expanduser().absolute()
STATIC_DIR: Final[Traversable] = DIR / "static"
TEMPLATES_DIR: Final[Traversable] = DIR / "templates"

ORJSON_OPTIONS: Final[int] = (
    json.OPT_SERIALIZE_NUMPY | json.OPT_NAIVE_UTC | json.OPT_UTC_Z
)

CONTAINERIZED: Final[bool] = "container" in os.environ or os.path.exists(
    "/.dockerenv"
)


def traversable_to_file(traversable: Traversable) -> Path:
    """Convert a traversable to a path to a file directly in the fs."""
    if isinstance(traversable, Path):
        return traversable
    folder = CACHE_DIR / "temp"
    folder.mkdir(parents=True, exist_ok=True)
    file = folder / traversable.name
    file.write_bytes(traversable.read_bytes())
    return file


CA_BUNDLE_PATH: Final[str] = traversable_to_file(
    DIR / "ca-bundle.crt"
).as_posix()


if pytest_is_running():
    NAME += "-test"
elif sys.flags.dev_mode:
    NAME += "-dev"

EVENT_SHUTDOWN: Final = multiprocessing.Event()

EVENT_ELASTICSEARCH: Final = Event()
EVENT_REDIS: Final = Event()

__all__ = (
    "CONTAINERIZED",
    "DIR",
    "EPOCH",
    "EPOCH_MS",
    "EVENT_ELASTICSEARCH",
    "EVENT_REDIS",
    "EVENT_SHUTDOWN",
    "GH_ORG_URL",
    "GH_PAGES_URL",
    "GH_REPO_URL",
    "MEDIA_TYPES",
    "NAME",
    "ORJSON_OPTIONS",
    "STATIC_DIR",
    "TEMPLATES_DIR",
    "UPTIME",
    "VERSION",
    "pytest_is_running",
)

__version__ = VERSION



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The Reporting API™️ of the website."""

from __future__ import annotations

import logging
from datetime import timedelta
from typing import Any, ClassVar, Final, cast

import orjson as json
from elasticsearch import AsyncElasticsearch
from elasticsearch.exceptions import NotFoundError
from elasticsearch.helpers import async_bulk
from tornado.web import HTTPError

from .. import EVENT_ELASTICSEARCH, ORJSON_OPTIONS
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo, Permission

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/reports", ReportingAPI),),
        name="Reporting API™️",
        description=(
            "Die Reporting API™️ kann zur Überwachung von "
            "Sicherheits-Verstößen, veralteten API-Aufrufen und mehr "
            "von Seiten des Asozialen Netzwerks genutzt werden.\n"
            "Bei Interesse kontakten Sie bitte das Gürteltier."
        ),
        path="/api/reports",
        hidden=True,
    )


async def get_reports(  # pylint: disable=too-many-arguments
    elasticsearch: AsyncElasticsearch,
    prefix: str,
    domain: None | str = None,
    type_: None | str = None,
    from_: int = 0,
    size: int = 10,
) -> list[dict[str, Any]]:
    """Get the reports from Elasticsearch."""
    query: dict[str, dict[str, list[dict[str, dict[str, Any]]]]]
    query = {"bool": {"filter": [{"range": {"@timestamp": {"gte": "now-1M"}}}]}}
    query["bool"]["must_not"] = [
        {
            "bool": {
                "filter": [
                    {"term": {"type": {"value": "network-error"}}},
                    {"term": {"body.type": {"value": "abandoned"}}},
                ]
            }
        },
        {
            "bool": {
                "filter": [
                    {"term": {"type": {"value": "csp-violation"}}},
                    {"term": {"body.source-file": {"value": "moz-extension"}}},
                ]
            }
        },
    ]
    if domain:
        query["bool"]["filter"].append(
            {
                "simple_query_string": {
                    "query": domain,
                    "fields": ["url.domain"],
                    "flags": "AND|ESCAPE|NOT|OR|PHRASE|PRECEDENCE|WHITESPACE",
                }
            }
        )
    if type_:
        query["bool"]["filter"].append(
            {
                "simple_query_string": {
                    "query": type_,
                    "fields": ["type"],
                    "flags": "AND|ESCAPE|NOT|OR|PHRASE|PRECEDENCE|WHITESPACE",
                }
            }
        )
    reports = await elasticsearch.search(
        index=f"{prefix}-reports",
        sort=[{"@timestamp": {"order": "desc"}}],
        query=query,
        from_=from_,
        size=size,
    )
    return [report["_source"] for report in reports["hits"]["hits"]]


class ReportingAPI(APIRequestHandler):
    """The request handler for the Reporting API™️."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        APIRequestHandler.POSSIBLE_CONTENT_TYPES + ("application/x-ndjson",)
    )

    RATELIMIT_GET_LIMIT: ClassVar[int] = 20
    RATELIMIT_GET_COUNT_PER_PERIOD: ClassVar[int] = 2

    RATELIMIT_POST_LIMIT: ClassVar[int] = 20
    RATELIMIT_POST_COUNT_PER_PERIOD: ClassVar[int] = 2

    MAX_BODY_SIZE: ClassVar[int] = 100_000_000

    MAX_REPORTS_PER_REQUEST: ClassVar[int] = 1000

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the Reporting API™️."""
        if not EVENT_ELASTICSEARCH.is_set():
            raise HTTPError(503)

        if head:
            return

        domain = self.get_argument("domain", None)
        type_ = self.get_argument("type", None)
        from_ = self.get_int_argument("from", 0, min_=0)
        size = self.get_int_argument("size", 10, min_=0)

        if not self.is_authorized(Permission.REPORTING):
            from_ = 0
            size = min(1000, size)

        try:
            reports = await get_reports(
                self.elasticsearch,
                self.elasticsearch_prefix,
                domain,
                type_,
                from_,
                size,
            )
        except NotFoundError:  # data stream doesn't exist
            raise HTTPError(404) from None

        if self.content_type == "application/x-ndjson":
            await self.finish(
                b"\n".join(
                    json.dumps(report, option=ORJSON_OPTIONS)
                    for report in reports
                )
            )
        else:
            await self.finish(self.dump(reports))

    async def post(self) -> None:
        """Handle POST requests to the Reporting API™️."""
        # pylint: disable=too-complex, too-many-branches
        if not (
            self.settings.get("REPORTING_BUILTIN")
            and EVENT_ELASTICSEARCH.is_set()
        ):
            raise HTTPError(503)
        if self.request.headers.get("Content-Type", "").startswith(
            "application/reports+json"
        ):
            reports = json.loads(self.request.body)
        elif self.request.headers.get("Content-Type", "").startswith(
            "application/csp-report"
        ):
            data = json.loads(self.request.body)
            if not isinstance(data, dict):
                raise HTTPError(400)
            body = data.get("csp-report")
            if not isinstance(body, dict):
                raise HTTPError(400)
            for camel, kebab in (
                ("blockedURL", "blocked-uri"),
                ("documentURL", "document-uri"),
                ("effectiveDirective", "effective-directive"),
                ("originalPolicy", "original-policy"),
                ("sample", "script-sample"),
                ("statusCode", "status-code"),
                ("violatedDirective", "violated-directive"),
            ):
                if kebab in body:
                    body[camel] = body.pop(kebab)  # 🥙 → 🐪
            report = {
                "age": 0,
                "body": body,
                "type": "csp-violation",
                "url": body.get("documentURL"),
                "user_agent": self.request.headers.get("User-Agent"),
            }
            reports = [report]
        else:
            raise HTTPError(415)
        if not isinstance(reports, list):
            raise HTTPError(400)
        if len(reports) > self.MAX_REPORTS_PER_REQUEST:
            LOGGER.warning(
                "%s > MAX_REPORTS_PER_REQUEST (%s)",
                len(reports),
                self.MAX_REPORTS_PER_REQUEST,
            )
            raise HTTPError(400)
        self.set_status(202)
        self.finish()  # type: ignore[unused-awaitable]
        for report in reports.copy():
            if not isinstance(report, dict):
                reports.remove(report)
                continue
            if isinstance((sauce := report.pop("_source", None)), dict):
                report.update(sauce)
            if not all(
                (
                    isinstance(report.get("age"), int),
                    isinstance(report.get("body"), dict),
                    isinstance(report.get("type"), str),
                    isinstance(report.get("url"), str),
                    isinstance(report.get("user_agent"), str),
                )
            ):
                reports.remove(report)
                continue
            report["@timestamp"] = self.now - timedelta(
                milliseconds=max(0, cast(int, report.pop("age")))
            )
            report["ecs"] = {"version": "8.17.0"}
            report["_op_type"] = "create"
            report.pop("_index", None)  # DO NOT REMOVE
        await async_bulk(
            self.elasticsearch,
            reports,
            index=f"{self.elasticsearch_prefix}-reports",
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The settings page used to change settings."""

from __future__ import annotations

import contextlib
import types
from base64 import b64encode
from collections.abc import Awaitable, Mapping

from tornado.web import HTTPError

from ..utils.options import COLOUR_SCHEMES, Options
from ..utils.request_handler import HTMLRequestHandler
from ..utils.themes import THEMES
from ..utils.utils import (
    BUMPSCOSITY_VALUES,
    ModuleInfo,
    bool_to_str,
    str_to_bool,
)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/einstellungen", SettingsPage),),
        name="Einstellungen",
        description="Stelle wichtige Sachen ein",
        path="/einstellungen",
        keywords=(
            "Einstellungen",
            "Config",
            "Settings",
        ),
        aliases=("/config", "/settings"),
    )


class SettingsPage(HTMLRequestHandler):
    """The request handler for the settings page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the settings page."""
        if head:
            return
        await self.render_settings(  # nosec: B106
            save_in_cookie=self.get_bool_argument("save_in_cookie", True),
            show_token_input=self.get_bool_argument("auth", False),
            token="",
            advanced_settings=self.show_advanced_settings(),
        )

    async def post(self) -> None:
        """Handle POST requests to the settings page."""
        if not self.request.body.strip():
            raise HTTPError(400, "No body provided")
        advanced_settings = self.get_bool_argument("advanced_settings", False)
        save_in_cookie: bool = self.get_bool_argument("save_in_cookie", False)
        token = self.get_argument("access_token", "")
        access_token: None | str = (
            b64encode(token.encode("UTF-8")).decode("ASCII") if token else None
        )

        if save_in_cookie:
            if access_token:
                self.set_cookie("access_token", access_token)
            self.set_cookie("advanced_settings", bool_to_str(advanced_settings))
            for option in self.user_settings.iter_options():
                self.set_cookie(
                    option.name,
                    option.value_to_string(
                        option.get_value(
                            self,
                            include_cookie=False,
                            include_query_argument=False,
                        )
                    ),
                    httponly=option.httponly,
                )

            replace_url_with = self.fix_url(
                self.request.full_url(),
                access_token=None,
                advanced_settings=None,
                save_in_cookie=None,
                **dict.fromkeys(self.user_settings.iter_option_names()),
            )
        else:
            replace_url_with = self.fix_url(
                self.request.full_url(),
                access_token=access_token,
                advanced_settings=advanced_settings,
                save_in_cookie=False,
                **self.user_settings.as_dict_with_str_values(
                    include_cookie=False, include_query_argument=False
                ),
            )

        if replace_url_with != self.request.full_url():
            return self.redirect(replace_url_with)

        await self.render_settings(
            advanced_settings=advanced_settings,
            save_in_cookie=save_in_cookie,
            show_token_input=bool(
                token or self.get_bool_argument("auth", False)
            ),
            token=token,
            kwargs={
                option.name: option.get_value(
                    self, include_cookie=False, include_query_argument=False
                )
                for option in self.user_settings.iter_options()
            },
        )

    def render_settings(  # pylint: disable=too-many-arguments
        self,
        *,
        save_in_cookie: bool,
        show_token_input: bool,
        token: str,
        advanced_settings: bool,
        kwargs: Mapping[str, object] = types.MappingProxyType({}),
    ) -> Awaitable[None]:
        """Render the settings page."""
        return self.render(
            "pages/settings.html",
            advanced_settings=advanced_settings,
            ask_before_leaving_default=(
                Options.ask_before_leaving.get_default_value(self)
            ),
            bumpscosity_values=BUMPSCOSITY_VALUES,
            no_3rd_party_default=Options.no_3rd_party.get_default_value(self),
            save_in_cookie=save_in_cookie,
            show_token_input=show_token_input,
            themes=THEMES,
            colour_schemes=COLOUR_SCHEMES,
            token=token,
            **kwargs,
        )

    def show_advanced_settings(self) -> bool:
        """Whether advanced settings should be shown."""
        if arg := self.get_argument("advanced_settings", ""):
            with contextlib.suppress(ValueError):
                return str_to_bool(arg)
        return str_to_bool(self.get_cookie("advanced_settings", ""), False)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
A page with wrong quotes.

It displays funny, but wrong, quotes.
"""

from __future__ import annotations

import asyncio
import logging
import random
from asyncio import AbstractEventLoop, Future
from dataclasses import dataclass
from typing import Any, ClassVar, Final, Literal, TypeAlias

import regex
from tornado.web import HTTPError

from .. import EVENT_REDIS
from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import hash_ip
from .image import IMAGE_CONTENT_TYPES, create_image
from .quote_of_the_day import QuoteOfTheDayBaseHandler
from .utils import (
    WRONG_QUOTES_CACHE,
    QuoteReadyCheckHandler,
    WrongQuote,
    create_wq_and_vote,
    get_authors,
    get_random_id,
    get_wrong_quote,
    get_wrong_quotes,
)

LOGGER: Final = logging.getLogger(__name__)


RatingFilter: TypeAlias = Literal["w", "n", "unrated", "rated", "all", "smart"]
SMART_RATING_FILTERS: Final[tuple[RatingFilter, ...]] = (
    *(("n",) * 1),
    *(("all",) * 5),
    *(("w",) * 5),
)


@dataclass(frozen=True)
class VoteArgument:
    """Voting matters."""

    vote: Literal[-1, 0, 1]


def parse_rating_filter(rating_filter_str: str) -> RatingFilter:
    """Get a rating filter."""
    match rating_filter_str:
        case "w":
            return "w"
        case "n":
            return "n"
        case "unrated":
            return "unrated"
        case "rated":
            return "rated"
        case "all":
            return "all"
    return "smart"


def get_next_id(rating_filter: RatingFilter) -> tuple[int, int]:
    """Get the id of the next quote."""
    if rating_filter == "smart":
        rating_filter = random.choice(SMART_RATING_FILTERS)  # nosec: B311

    match rating_filter:
        case "unrated":
            # get a random quote, but filter out already rated quotes
            # pylint: disable=while-used
            while (ids := get_random_id()) in WRONG_QUOTES_CACHE:
                if WRONG_QUOTES_CACHE[ids].id == -1:
                    # check for wrong quotes that are unrated but in the cache
                    # they don't have a real wrong_quotes_id
                    return ids
            return ids
        case "all":
            return get_random_id()
        case "w":
            wrong_quotes = get_wrong_quotes(lambda wq: wq.rating > 0)
        case "n":
            wrong_quotes = get_wrong_quotes(lambda wq: wq.rating < 0)
        case "rated":
            wrong_quotes = get_wrong_quotes(lambda wq: wq.id not in {-1, None})
        case _:
            LOGGER.error("Invalid rating filter %s", rating_filter)
            return get_random_id()

    if not wrong_quotes:
        # invalid rating filter or no wrong quotes with that filter
        return get_random_id()

    return random.choice(wrong_quotes).get_id()  # nosec: B311


class QuoteBaseHandler(QuoteReadyCheckHandler):
    """The base request handler for the quotes package."""

    RATELIMIT_GET_LIMIT = 20
    RATELIMIT_GET_COUNT_PER_PERIOD = 20
    RATELIMIT_GET_PERIOD = 10

    FUTURES: set[Future[Any]] = set()

    loop: AbstractEventLoop
    next_id: tuple[int, int]
    rating_filter: RatingFilter

    def future_callback(self, future: Future[WrongQuote]) -> None:
        """Discard the future and log the exception if one occured."""
        self.FUTURES.discard(future)
        if exc := future.exception():
            LOGGER.error(
                "Failed to pre-fetch quote %d-%d (%s)",
                *self.next_id,
                exc,
                exc_info=(type(exc), exc, exc.__traceback__),
            )
        else:
            LOGGER.debug("Pre-fetched quote %d-%d", *self.next_id)

    def get_next_url(self) -> str:
        """Get the URL of the next quote."""
        return self.fix_url(
            f"/zitate/{self.next_id[0]}-{self.next_id[1]}",
            r=None if self.rating_filter == "smart" else self.rating_filter,
            **{"show-rating": self.get_show_rating() or None},  # type: ignore[arg-type]
        )

    def get_show_rating(self) -> bool:
        """Return whether the user wants to see the rating."""
        return self.get_bool_argument("show-rating", False)

    def on_finish(self) -> None:
        """
        Pre-fetch the data for the next quote.

        This is done to show the users less out-of-date data.
        """
        if len(self.FUTURES) > 1 or (
            self.content_type
            and self.content_type.startswith("image/")
            or self.content_type
            in {"application/pdf", "application/vnd.ms-excel"}
        ):
            return  # don't spam and don't do this for images

        user_agent = self.request.headers.get("User-Agent")
        if not user_agent or "bot" in user_agent.lower():
            return  # don't do this for bots

        if hasattr(self, "loop"):
            task = self.loop.create_task(
                get_wrong_quote(*self.next_id, use_cache=False)
            )
            self.FUTURES.add(task)
            task.add_done_callback(self.future_callback)

    async def prepare(self) -> None:
        """Set the id of the next wrong_quote to show."""
        await super().prepare()
        self.loop = asyncio.get_running_loop()
        self.rating_filter = parse_rating_filter(self.get_argument("r", ""))
        self.next_id = get_next_id(self.rating_filter)


class QuoteMainPage(QuoteBaseHandler, QuoteOfTheDayBaseHandler):
    """The main quote page that explains everything and links to stuff."""

    async def get(self, *, head: bool = False) -> None:
        """Render the main quote page, with a few links."""
        # pylint: disable=unused-argument
        quote = self.get_argument("quote", "")
        author = self.get_argument("author", "")
        if (quote or author) and regex.fullmatch(r"^[0-9]+$", quote + author):
            self.redirect(self.fix_url(f"/zitate/{quote}-{author}"))
            return

        await self.render(
            "pages/quotes/main_page.html",
            funny_quote_url=self.id_to_url(
                *(
                    get_wrong_quotes(lambda wq: wq.rating > 0, shuffle=True)[
                        0
                    ].get_id()
                ),
                rating_param="w",
            ),
            random_quote_url=self.id_to_url(*self.next_id),
            quote_of_the_day=self.redis and await self.get_quote_of_today(),  # type: ignore[truthy-bool]
            one_stone_url=self.get_author_url("Albert Einstein"),
            kangaroo_url=self.get_author_url("Das Känguru"),
            muk_url=self.get_author_url("Marc-Uwe Kling"),
        )

    def get_author_url(self, author_name: str) -> None | str:
        """Get the info URL of an author."""
        authors = get_authors(lambda _a: _a.name.lower() == author_name.lower())
        if not authors:
            return None
        return self.fix_url(f"/zitate/info/a/{authors[0].id}")

    def id_to_url(
        self, quote_id: int, author_id: int, rating_param: None | str = None
    ) -> str:
        """Get the URL of a quote."""
        return self.fix_url(f"/zitate/{quote_id}-{author_id}", r=rating_param)


def wrong_quote_to_json(
    wq_: WrongQuote, vote: int, next_: tuple[int, int], rating: str, full: bool
) -> dict[str, Any]:  # TODO: improve lazy return type typing
    """Convert a wrong quote to a dict."""
    next_q, next_a = next_
    if full:
        return {
            "wrong_quote": wq_.to_json(),
            "next": f"{next_q}-{next_a}",
            "vote": vote,
        }
    return {
        "id": wq_.get_id_as_str(),
        "quote": str(wq_.quote),
        "author": str(wq_.author),
        "real_author": str(wq_.quote.author),
        "real_author_id": wq_.quote.author_id,
        "rating": rating,
        "vote": vote,
        "next": f"{next_q}-{next_a}",
    }


class QuoteById(QuoteBaseHandler):
    """The page with a specified quote that then gets rendered."""

    RATELIMIT_POST_LIMIT: ClassVar[int] = 10
    RATELIMIT_POST_COUNT_PER_PERIOD: ClassVar[int] = 5
    RATELIMIT_POST_PERIOD: ClassVar[int] = 10

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        *HTMLRequestHandler.POSSIBLE_CONTENT_TYPES,
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
        *IMAGE_CONTENT_TYPES,
    )

    LONG_PATH: ClassVar[str] = "/zitate/%d-%d"

    async def get(
        self, quote_id: str, author_id: None | str = None, *, head: bool = False
    ) -> None:
        """Handle GET requests to this page and render the quote."""
        int_quote_id = int(quote_id)
        if author_id is None:
            wqs = get_wrong_quotes(lambda wq: wq.id == int_quote_id)
            if not wqs:
                raise HTTPError(404, f"No wrong quote with id {quote_id}")
            return self.redirect(self.fix_url(self.LONG_PATH % wqs[0].get_id()))

        if head:
            return

        if (
            self.content_type
            and self.content_type.startswith("image/")
            or self.content_type
            in {"application/pdf", "application/vnd.ms-excel"}
        ):
            wrong_quote = await get_wrong_quote(int_quote_id, int(author_id))
            return await self.finish(
                await asyncio.to_thread(
                    create_image,
                    wrong_quote.quote.quote,
                    wrong_quote.author.name,
                    wrong_quote.rating,
                    f"{self.request.host_name}/z/{wrong_quote.get_id_as_str(True)}",
                    (
                        self.content_type.removeprefix("image/").removeprefix(
                            "x-"
                        )
                        if self.content_type.startswith("image/")
                        else {
                            "application/pdf": "pdf",
                            "application/vnd.ms-excel": "xlsx",
                        }[self.content_type]
                    ),
                    wq_id=wrong_quote.get_id_as_str(),
                )
            )

        await self.render_quote(int_quote_id, int(author_id))

    async def get_old_vote(
        self, quote_id: int, author_id: int
    ) -> Literal[-1, 0, 1]:
        """Get the old vote from the saved vote."""
        old_vote = await self.get_saved_vote(quote_id, author_id)
        if old_vote is None:
            return 0
        return old_vote

    async def get_rating_str(self, wrong_quote: WrongQuote) -> str:
        """Get the rating str to display on the page."""
        if wrong_quote.id in {None, -1}:
            return "---"
        if (
            not self.get_show_rating()  # don't hide the rating on wish of user
            and self.rating_filter == "smart"
            and self.request.method
            and self.request.method.upper() == "GET"
            and await self.get_saved_vote(
                wrong_quote.quote_id, wrong_quote.author_id
            )
            is None
        ):
            return "???"
        return str(wrong_quote.rating)

    def get_redis_votes_key(self, quote_id: int, author_id: int) -> str:
        """Get the key to save the votes with Redis."""
        return (
            f"{self.redis_prefix}:quote-votes:"
            f"{self.get_user_id()}:{quote_id}-{author_id}"
        )

    async def get_saved_vote(
        self, quote_id: int, author_id: int
    ) -> None | Literal[-1, 0, 1]:
        """
        Get the vote of the current user saved with Redis.

        Use the quote_id and author_id to query the vote.
        Return None if nothing is saved.
        """
        if not EVENT_REDIS.is_set():
            LOGGER.warning("No Redis connection")
            return 0
        result = await self.redis.get(
            self.get_redis_votes_key(quote_id, author_id)
        )
        if result == "-1":
            return -1
        if result == "0":
            return 0
        if result == "1":
            return 1
        return None

    @parse_args(type_=VoteArgument)
    async def post(
        self,
        quote_id_str: str,
        author_id_str: str | None = None,
        *,
        args: VoteArgument,
    ) -> None:
        """
        Handle POST requests to this page and render the quote.

        This is used to vote the quote, without changing the URL.
        """
        quote_id = int(quote_id_str)
        if author_id_str is None:
            wqs = get_wrong_quotes(lambda wq: wq.id == quote_id)
            if not wqs:
                raise HTTPError(404, f"No wrong quote with id {quote_id}")
            return self.redirect(self.fix_url(self.LONG_PATH % wqs[0].get_id()))

        author_id = int(author_id_str)

        new_vote_str = self.get_argument("vote", None)

        if not new_vote_str:
            return await self.render_quote(quote_id, author_id)

        old_vote: Literal[-1, 0, 1] = await self.get_old_vote(
            quote_id, author_id
        )

        new_vote: Literal[-1, 0, 1] = args.vote
        vote_diff: int = new_vote - old_vote

        if not vote_diff:  # == 0
            return await self.render_quote(quote_id, author_id)

        await self.update_saved_votes(quote_id, author_id, new_vote)

        to_vote: Literal[-1, 1] = -1 if vote_diff < 0 else 1

        contributed_by = f"an-website_{hash_ip(self.request.remote_ip, 10)}"

        # do the voting
        wrong_quote = await create_wq_and_vote(
            to_vote, quote_id, author_id, contributed_by
        )
        if abs(vote_diff) == 2:
            await wrong_quote.vote(to_vote, lazy=True)

        await self.render_wrong_quote(wrong_quote, new_vote)

    async def render_quote(self, quote_id: int, author_id: int) -> None:
        """Get and render a wrong quote based on author id and author id."""
        await self.render_wrong_quote(
            await get_wrong_quote(quote_id, author_id),
            await self.get_old_vote(quote_id, author_id),
        )

    async def render_wrong_quote(
        self, wrong_quote: WrongQuote, vote: int
    ) -> None:
        """Render the page with the wrong_quote and this vote."""
        if self.content_type in APIRequestHandler.POSSIBLE_CONTENT_TYPES:
            return await self.finish(
                wrong_quote_to_json(
                    wrong_quote,
                    vote,
                    self.next_id,
                    await self.get_rating_str(wrong_quote),
                    self.get_bool_argument("full", False),
                )
            )
        await self.render(
            "pages/quotes/quotes.html",
            wrong_quote=wrong_quote,
            next_href=self.get_next_url(),
            next_id=f"{self.next_id[0]}-{self.next_id[1]}",
            description=str(wrong_quote),
            rating_filter=self.rating_filter,
            rating=await self.get_rating_str(wrong_quote),
            show_rating=self.get_show_rating(),
            vote=vote,
        )

    async def update_saved_votes(
        self, quote_id: int, author_id: int, vote: int
    ) -> None:
        """Save the new vote in Redis."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)
        result = await self.redis.setex(
            self.get_redis_votes_key(quote_id, author_id),
            60 * 60 * 24 * 90,  # time to live in seconds (3 months)
            str(vote),  # value to save (the vote)
        )
        if result:
            return
        LOGGER.warning("Could not save vote in Redis: %s", result)
        raise HTTPError(500, "Could not save vote")


# pylint: disable-next=too-many-ancestors
class QuoteAPIHandler(APIRequestHandler, QuoteById):
    """API request handler for the quotes page."""

    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("GET", "POST")

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
        *IMAGE_CONTENT_TYPES,
    )

    LONG_PATH: ClassVar[str] = "/api/zitate/%d-%d"

    async def render_wrong_quote(
        self, wrong_quote: WrongQuote, vote: int
    ) -> None:
        """Return the relevant data for the quotes page as JSON."""
        if self.content_type == "text/plain":
            return await self.finish(str(wrong_quote))
        return await self.finish(
            wrong_quote_to_json(
                wrong_quote,
                vote,
                self.next_id,
                await self.get_rating_str(wrong_quote),
                self.request.path.endswith("/full"),
            )
        )


class QuoteRedirectAPI(APIRequestHandler, QuoteBaseHandler):
    """Redirect to the API for a random quote."""

    POSSIBLE_CONTENT_TYPES = QuoteAPIHandler.POSSIBLE_CONTENT_TYPES

    async def get(  # pylint: disable=unused-argument
        self, suffix: str = "", *, head: bool = False
    ) -> None:
        """Redirect to a random funny quote."""
        next_filter = parse_rating_filter(self.get_argument("r", "") or "w")
        quote_id, author_id = get_next_id(next_filter)
        kwargs: dict[str, str] = {"r": next_filter}
        if self.get_show_rating():
            kwargs["show-rating"] = "sure"
        return self.redirect(
            self.fix_url(
                f"/api/zitate/{quote_id}-{author_id}{suffix}",
                **kwargs,
            )
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test Christmas."""

from __future__ import annotations

import time_machine

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_html_response,
    fetch,
)


@time_machine.travel("1961-12-15")
async def test_theme_on_main_page(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the random text API."""
    response = assert_valid_html_response(
        await fetch("/200.html?theme=default"), codes={200}
    )
    assert b"christmas.css" in response.body
    assert b"default.css" not in response.body



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The soundboard of the website."""

from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import cache
from typing import ClassVar

from tornado.web import HTTPError

from ..utils.request_handler import HTMLRequestHandler
from .data import (
    ALL_SOUNDS,
    MAIN_PAGE_INFO,
    PERSON_SHORTS,
    PERSON_SOUNDS,
    HeaderInfo,
    Info,
    Person,
    SoundInfo,
)


@cache
def get_rss_str(path: str, protocol_and_host: str) -> None | str:
    """Return the RSS string for the given path."""
    if path is not None:
        path = path.lower()

    if path in {None, "/", ""}:
        _infos: Iterable[SoundInfo] = ALL_SOUNDS
    elif path in PERSON_SOUNDS:
        _infos = PERSON_SOUNDS[path]
    else:
        return None
    return "\n".join(
        sound_info.to_rss(protocol_and_host) for sound_info in _infos
    )


async def search_main_page_info(
    check_func: Callable[[SoundInfo], bool],
    info_list: Iterable[Info] = MAIN_PAGE_INFO,
) -> list[Info]:
    # pylint: disable=confusing-consecutive-elif
    """Get an info list based on the query and the check_func and return it."""
    found: list[Info] = []
    for info in info_list:
        if isinstance(info, SoundInfo):
            if check_func(info):
                found.append(info)
        elif isinstance(info, HeaderInfo):
            tag = info.tag
            while (  # pylint: disable=while-used
                len(found) > 0
                and isinstance(last := found[-1], HeaderInfo)
                and (
                    tag
                    in (
                        "h1",  # if it gets to h3 this doesn't work as
                        # then this should also be done for h2 when the ones
                        # before are h3
                        last.tag,
                    )
                )
            ):
                del found[-1]
            found.append(info)

    # pylint: disable=while-used
    while len(found) > 0 and isinstance(found[-1], HeaderInfo):
        del found[-1]

    return found


class SoundboardHTMLHandler(HTMLRequestHandler):
    """The request handler for the HTML pages."""

    async def get(self, path: str = "/", *, head: bool = False) -> None:
        """Handle GET requests and generate the page content."""
        if path is not None:
            path = path.lower()

        parsed_info = await self.parse_path(path)
        if parsed_info is None:
            raise HTTPError(404, reason="Page not found")

        if head:
            return

        self.update_title_and_desc(path)

        await self.render(
            "pages/soundboard.html",
            sound_info_list=parsed_info[0],
            query=parsed_info[1],
            feed_url=self.fix_url(
                (
                    f"/soundboard/{path.strip('/')}/feed"
                    if path and path != "/" and path != "personen"
                    else "/soundboard/feed"
                ),
            ),
        )

    async def parse_path(
        self, path: None | str
    ) -> None | tuple[Iterable[Info], None | str]:
        """Get an info list based on the path and return it with the query."""
        if path in {None, "", "index", "/"}:
            return MAIN_PAGE_INFO, None

        if path in {"persons", "personen"}:
            persons_list: list[Info] = []
            for key, person_sounds in PERSON_SOUNDS.items():
                persons_list.append(HeaderInfo(Person[key].value, type=Person))
                persons_list += person_sounds
            return persons_list, None

        if path in {"search", "suche"}:
            query = self.get_argument("q", "")
            if not query:
                return MAIN_PAGE_INFO, query

            return (
                await search_main_page_info(lambda info: info.contains(query)),
                query,
            )

        if path in PERSON_SHORTS:
            person = Person[path]
            return (
                await search_main_page_info(lambda info: info.person == person),
                None,
            )

        return None

    def update_title_and_desc(self, path: str) -> None:
        """Update the title and description of the page."""
        if path not in PERSON_SHORTS:
            return
        name = Person[path].value
        if name.startswith("Das ") or name.startswith("Der "):
            von_name = f"dem{name[3:]}"
            no_article_name = name[4:]
        elif name.startswith("Die "):
            von_name = f"der{name[3:]}"
            no_article_name = name[4:]
        else:
            von_name = name
            no_article_name = name

        self.short_title = f"Soundboard ({path.upper()})"
        self.title = f"{no_article_name.replace(' ', '-')}-Soundboard"
        self.description = (
            "Ein Soundboard mit coolen Sprüchen und Sounds von "
            f"{von_name} aus den Känguru-Chroniken"
        )


class SoundboardRSSHandler(SoundboardHTMLHandler):
    """The request handler for the RSS feeds."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "application/rss+xml",
        "application/xml",
    )

    async def get(self, path: str = "/", *, head: bool = False) -> None:
        """Handle GET requests and generate the feed content."""
        rss_str = get_rss_str(
            path, f"{self.request.protocol}://{self.request.host}"
        )

        if rss_str is not None:
            if head:
                return
            self.update_title_and_desc(path)
            return await self.render(
                "rss/soundboard.xml",
                found=True,
                rss_str=rss_str,
            )
        self.set_status(404, reason="Feed not found")
        return await self.render("rss/soundboard.xml", found=False, rss_str="")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Create the required stuff for the soundboard from the info in info.json."""

from __future__ import annotations

import email.utils
from collections.abc import Callable
from dataclasses import dataclass, field
from enum import Enum
from typing import Final

import orjson as json
import regex

from .. import DIR as ROOT_DIR
from ..utils.fix_static_path_impl import hash_file
from ..utils.utils import name_to_id, replace_umlauts, size_of_file

DIR: Final = ROOT_DIR / "soundboard"
with (DIR / "info.json").open("r", encoding="UTF-8") as file:
    info = json.loads(file.read())

# {"muk": "Marc-Uwe Kling", ...}
person_dict: dict[str, str] = info["personen"]
Person = Enum("Person", {**person_dict}, module=__name__)  # type: ignore[misc]

PERSON_SHORTS: tuple[str, ...] = tuple(person_dict.keys())

books: list[str] = []
chapters: list[str] = []
for book in info["bücher"]:
    books.append(book["name"])

    for chapter in book["kapitel"]:
        chapters.append(chapter["name"])


Book = Enum("Book", [*books], module=__name__)  # type: ignore[misc]
Chapter = Enum("Chapter", [*chapters], module=__name__)  # type: ignore[misc]


del books, chapters, person_dict


def mark_query(text: str, query: None | str) -> str:
    """Replace the instances of the query with itself in a div."""
    if not query:
        return text

    query = regex.sub("(ä|ae)", "(ä|ae)", query.lower())
    query = regex.sub("(ö|oe)", "(ö|oe)", query)
    query = regex.sub("(ü|ue)", "(ü|ue)", query)
    query = regex.sub("(ü|ue)", "(ü|ue)", query)

    for word in query.split(" "):
        text = regex.sub(
            word,
            lambda match: f'<div class="marked">{match[0]}</div>',  # type: ignore[str-bytes-safe]  # noqa: B950
            text,
            regex.IGNORECASE,
        )

    return text


@dataclass(frozen=True, slots=True)
class Info:
    """Info class that is used as a base for HeaderInfo and SoundInfo."""

    text: str

    def to_html(
        self,
        fix_url_func: Callable[  # pylint: disable=unused-argument
            [str], str
        ] = lambda url: url,
        query: None | str = None,
    ) -> str:
        """Return the text of the info and mark the query."""
        return mark_query(self.text, query)


@dataclass(frozen=True, slots=True)
class HeaderInfo(Info):
    """A header with a tag and href to itself."""

    tag: str = "h1"
    type: type[Book | Chapter | Person] = Book

    def to_html(
        self,
        fix_url_func: Callable[[str], str] = lambda url: url,
        query: None | str = None,
    ) -> str:
        """
        Return an HTML element with the tag and the content of the HeaderInfo.

        The HTML element gets an id and a href
        with a # to itself based on the text content.
        """
        id_ = name_to_id(self.text)
        text = (
            self.text  # no need to mark query if type
            # is  book as the book title is excluded from the search
            if self.type == Book
            else mark_query(self.text, query)
        )
        return (
            f"<{self.tag} id={id_!r}>"
            f"{text}<a no-dynload href='#{id_}' class='header-id-link'></a>"
            f"</{self.tag}>"
        )


@dataclass(frozen=True, slots=True)
class SoundInfo(Info):
    """The information about a sound."""

    book: Book
    chapter: Chapter
    person: Person

    filename: str = field(init=False)
    pub_date: str = field(init=False)

    def __post_init__(self) -> None:
        """Init post."""
        person_, timestamp, text = self.text.split("-", 2)
        object.__setattr__(
            self,
            "filename",
            regex.sub(
                r"[^a-z0-9_-]+",
                "",
                replace_umlauts(
                    (person_ + "-" + text).lower().replace(" ", "_")
                ),
            ),
        )
        object.__setattr__(self, "text", text)
        # convert seconds since epoch to readable timestamp
        object.__setattr__(
            self, "pub_date", email.utils.formatdate(float(timestamp), True)
        )

    def contains(self, string: None | str) -> bool:
        """Check whether this sound info contains a given string."""
        if string is None:
            return False

        content = " ".join([self.chapter.name, self.person.value, self.text])
        content = replace_umlauts(content.lower())

        return not any(
            word not in content
            for word in replace_umlauts(string.lower()).split(" ")
        )

    def to_html(
        self,
        fix_url_func: Callable[[str], str] = lambda url: url,
        query: None | str = None,
    ) -> str:
        """Parse the info to a list element with an audio element."""
        file = self.filename  # pylint: disable=redefined-outer-name
        href = fix_url_func(f"/soundboard/{self.person.name}")
        path = f"files/{file}.mp3"
        file_url = f"/soundboard/{path}?v={hash_file(DIR / path)}"
        return (
            f"<li id={file!r}>"
            f"<a href={href!r} class='a_hover'>"
            f"{mark_query(self.person.value, query)}"
            "</a>"
            ": »"
            f"<a no-dynload href={file_url!r} class='quote-a'>"
            f"{mark_query(self.text, query)}"
            "</a>"
            "«"
            "<br>"
            '<audio controls preload="none">'
            f"<source src={file_url!r} type='audio/mpeg'>"
            "</audio>"
            "</li>"
        )

    def to_rss(self, url: None | str) -> str:
        """Parse the info to a RSS item."""
        filename = self.filename
        rel_path = f"files/{filename}.mp3"
        abs_path = DIR / rel_path
        file_size = size_of_file(abs_path)
        link = f"/soundboard/{rel_path}"
        text = self.text
        if url is not None:
            link = url + link
        return (
            "<item>"
            "<title>"
            f"[{self.book.name} - {self.chapter.name}] "
            f"{self.person.value}: »{text}«"
            "</title>"
            f"<quote>{text}</quote>"
            f"<book>{self.book.name}</book>"
            f"<chapter>{self.chapter.name}</chapter>"
            f"<person>{self.person.value}</person>"
            f"<link>{link}</link>"
            f"<enclosure url={link!r} type='audio/mpeg'"
            f" length={str(file_size)!r}>"
            "</enclosure>"
            f"<guid>{filename}</guid>"
            f"<pubDate>{self.pub_date}</pubDate>"
            "</item>"
        )


all_sounds: list[SoundInfo] = []
main_page_info: list[Info] = []
PERSON_SOUNDS: Final[dict[str, list[SoundInfo]]] = {}

for book_info in info["bücher"]:
    book = Book[book_info["name"]]
    main_page_info.append(HeaderInfo(book.name, "h1", Book))

    for chapter_info in book_info["kapitel"]:
        chapter = Chapter[chapter_info["name"]]
        main_page_info.append(HeaderInfo(chapter.name, "h2", Chapter))

        for file_text in chapter_info["dateien"]:
            person_short = file_text.split("-")[0]
            person = Person[person_short]

            sound_info = SoundInfo(file_text, book, chapter, person)
            all_sounds.append(sound_info)
            main_page_info.append(sound_info)
            PERSON_SOUNDS.setdefault(person_short, []).append(sound_info)

# convert to tuple for immutability
ALL_SOUNDS: Final[tuple[SoundInfo, ...]] = tuple(all_sounds)
MAIN_PAGE_INFO: Final[tuple[Info, ...]] = tuple(main_page_info)

del all_sounds, main_page_info



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A config parser with the ability to parse command-line arguments."""
from __future__ import annotations

import argparse
import logging
import pathlib
from collections.abc import Callable, Iterable
from configparser import ConfigParser
from typing import Any, Final, TypeVar, cast, overload

from .utils import bool_to_str, get_arguments_without_help, str_to_set

LOGGER: Final = logging.getLogger(__name__)

T = TypeVar("T")
OptionalBool = TypeVar("OptionalBool", bool, None)
OptionalFloat = TypeVar("OptionalFloat", float, None)
OptionalInt = TypeVar("OptionalInt", int, None)
OptionalStr = TypeVar("OptionalStr", str, None)
OptionalSetStr = TypeVar("OptionalSetStr", set[str], None)


class BetterConfigParser(ConfigParser):
    """A better config parser."""

    _arg_parser: None | argparse.ArgumentParser
    _arg_parser_options_added: set[tuple[str, str]]
    _all_options_should_be_parsed: bool

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Initialize this config parser."""
        self._arg_parser_options_added = set()
        self._arg_parser = None
        self._all_options_should_be_parsed = False
        kwargs.setdefault("interpolation", None)
        kwargs["dict_type"] = dict
        super().__init__(*args, **kwargs)

    def _add_fallback_to_config(
        self,
        section: str,
        option: str,
        fallback: str | Iterable[str] | bool | int | float | None,
    ) -> None:
        if section in self and option in self[section]:
            return
        fallback = self._val_to_str(fallback)
        if fallback is None:
            fallback = ""
            option = f"#{option}"
        if section not in self.sections():
            self.add_section(section)
        self.set(section, option.lower(), fallback)

    def _get_conv(  # type: ignore[override]
        self, section: str, option: str, conv: Callable[[str], T], **kwargs: Any
    ) -> T | None:
        self._add_fallback_to_config(section, option, kwargs.get("fallback"))
        if (val := self._get_from_args(section, option, conv)) is not None:
            return val
        return cast(T, super()._get_conv(section, option, conv, **kwargs))

    def _get_from_args(
        self, section: str, option: str, conv: Callable[[str], T]
    ) -> None | T:
        """Try to get the value from the command line arguments."""
        if self._arg_parser is None:
            return None
        option_name = f"{section}-{option}".lower().removeprefix("general-")
        if (section, option) not in self._arg_parser_options_added:
            if self._all_options_should_be_parsed:
                LOGGER.error(
                    "Option %r in section %r should have been queried before.",
                    option,
                    section,
                )
            self._arg_parser.add_argument(
                f"--{option_name}".replace("_", "-"),
                required=False,
                type=conv,
                help=f"Override {option!r} in the {section!r} section of the config",
            )
            self._arg_parser_options_added.add((section, option))
        value = getattr(
            self._arg_parser.parse_known_args(get_arguments_without_help())[0],
            option_name.replace("-", "_"),
            None,
        )
        if value is None:
            return None
        self.set(section, option, self._val_to_str(value))
        return cast(T, value)

    def _val_to_str(self, value: object | None) -> str | None:
        """Convert a value to a string."""
        if value is None or isinstance(value, str):
            return value
        if isinstance(value, Iterable):
            return ", ".join(
                [cast(str, self._val_to_str(val)) for val in value]
            )
        if isinstance(value, bool):
            return bool_to_str(value)
        return str(value)  # float, int

    def add_override_argument_parser(
        self, parser: argparse.ArgumentParser
    ) -> None:
        """Add an argument parser to override config values."""
        self._arg_parser = parser

    @staticmethod
    def from_path(*path: pathlib.Path) -> BetterConfigParser:
        """Parse the config at the given path."""
        config = BetterConfigParser()
        config.read(path, encoding="UTF-8")
        return config

    @overload  # type: ignore[override]
    def get(  # noqa: D102  # pylint: disable=arguments-differ
        self, section: str, option: str, *, fallback: OptionalStr
    ) -> str | OptionalStr: ...

    @overload
    def get(  # pylint: disable=arguments-differ
        self, section: str, option: str
    ) -> str: ...

    def get(self, section: str, option: str, **kwargs: Any) -> object:
        """Get an option in a section."""
        self._add_fallback_to_config(section, option, kwargs.get("fallback"))
        if (val := self._get_from_args(section, option, str)) is not None:
            assert isinstance(val, str)
            return val
        return_value: str | None = super().get(section, option, **kwargs)
        if "fallback" in kwargs and kwargs["fallback"] is None:
            assert isinstance(return_value, (str, type(None)))
        else:
            assert isinstance(return_value, str)
        return return_value

    @overload  # type: ignore[override]
    def getboolean(  # noqa: D102  # pylint: disable=arguments-differ
        self, section: str, option: str, *, fallback: OptionalBool
    ) -> bool | OptionalBool: ...

    @overload
    def getboolean(  # pylint: disable=arguments-differ
        self, section: str, option: str
    ) -> bool: ...

    def getboolean(self, section: str, option: str, **kwargs: Any) -> object:
        """Get a boolean option."""
        return_value: bool | None = super().getboolean(
            section, option, **kwargs
        )
        if "fallback" in kwargs and kwargs["fallback"] is None:
            assert return_value in {False, None, True}
        else:
            assert return_value in {False, True}
        return return_value

    @overload  # type: ignore[override]
    def getfloat(  # noqa: D102  # pylint: disable=arguments-differ
        self, section: str, option: str, *, fallback: OptionalFloat
    ) -> float | OptionalFloat: ...

    @overload
    def getfloat(  # pylint: disable=arguments-differ
        self, section: str, option: str
    ) -> float: ...

    def getfloat(self, section: str, option: str, **kwargs: Any) -> object:
        """Get an int option."""
        return_value: float | None = super().getfloat(section, option, **kwargs)
        if "fallback" in kwargs and kwargs["fallback"] is None:
            assert isinstance(return_value, (float, type(None)))
        else:
            assert isinstance(return_value, float)
        return return_value

    @overload  # type: ignore[override]
    def getint(  # noqa: D102  # pylint: disable=arguments-differ
        self, section: str, option: str, *, fallback: OptionalInt
    ) -> int | OptionalInt: ...

    @overload
    def getint(  # pylint: disable=arguments-differ
        self, section: str, option: str
    ) -> int: ...

    def getint(self, section: str, option: str, **kwargs: Any) -> object:
        """Get an int option."""
        return_value: int | None = super().getint(section, option, **kwargs)
        if "fallback" in kwargs and kwargs["fallback"] is None:
            assert isinstance(return_value, (int, type(None)))
        else:
            assert isinstance(return_value, int)
        return return_value

    @overload
    def getset(  # noqa: D102
        self, section: str, option: str, *, fallback: OptionalSetStr
    ) -> set[str] | OptionalSetStr: ...

    @overload
    def getset(self, section: str, option: str) -> set[str]: ...

    def getset(self, section: str, option: str, **kwargs: Any) -> object:
        """Get an int option."""
        return_value: set[str] | None = self._get_conv(
            section, option, str_to_set, **kwargs
        )
        if "fallback" in kwargs and kwargs["fallback"] is None:
            assert isinstance(return_value, (set, type(None)))
        else:
            assert isinstance(return_value, set)
        if isinstance(return_value, set):
            assert all(isinstance(val, str) for val in return_value)
        return return_value

    def set_all_options_should_be_parsed(self) -> None:
        """Set all options should be parsed."""
        self._all_options_should_be_parsed = True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Get a random quote for a given day."""

from __future__ import annotations

import logging
from datetime import date, datetime, timedelta, timezone
from typing import ClassVar, Final

from tornado.web import HTTPError

from ...utils.request_handler import APIRequestHandler
from ..utils import (
    QuoteReadyCheckHandler,
    WrongQuote,
    get_wrong_quote,
    get_wrong_quotes,
)
from .data import QuoteOfTheDayData
from .store import (
    QUOTE_COUNT_TO_SHOW_IN_FEED,
    QuoteOfTheDayStore,
    RedisQuoteOfTheDayStore,
)

LOGGER: Final = logging.getLogger(__name__)


class QuoteOfTheDayBaseHandler(QuoteReadyCheckHandler):
    """The base request handler for the quote of the day."""

    async def get_quote_by_date(
        self, wq_date: date | str
    ) -> None | QuoteOfTheDayData:
        """Get the quote of the date if one was saved."""
        if isinstance(wq_date, str):
            wq_date = date.fromisoformat(wq_date)

        wq_id = await self.qod_store.get_quote_id_by_date(wq_date)
        if not wq_id:
            return None
        wrong_quote = await get_wrong_quote(*wq_id)
        if not wrong_quote:  # type: ignore[truthy-bool]
            return None
        return QuoteOfTheDayData(
            wq_date, wrong_quote, self.get_scheme_and_netloc()
        )

    async def get_quote_of_today(self) -> None | QuoteOfTheDayData:
        """Get the quote for today."""
        today = datetime.now(timezone.utc).date()
        quote_data = await self.get_quote_by_date(today)
        if quote_data:  # if was saved already
            return quote_data
        quotes = get_wrong_quotes(lambda wq: wq.rating > 1, shuffle=True)
        if not quotes:
            LOGGER.error("No quotes available")
            return None
        for quote in quotes:
            if await self.qod_store.has_quote_been_used(quote.get_id()):
                continue
            wq_id = quote.get_id()
            await self.qod_store.set_quote_to_used(wq_id)
            await self.qod_store.set_quote_id_by_date(today, wq_id)
            return QuoteOfTheDayData(today, quote, self.get_scheme_and_netloc())
        LOGGER.critical("Failed to generate a new quote of the day")
        return None

    def get_scheme_and_netloc(self) -> str:
        """Get the beginning of the URL."""
        return f"{self.request.protocol}://{self.request.host}"

    @property
    def qod_store(self) -> QuoteOfTheDayStore:
        """Get the store used for storing the quote of the day."""
        return RedisQuoteOfTheDayStore(self.redis, self.redis_prefix)


class QuoteOfTheDayRSS(QuoteOfTheDayBaseHandler):
    """The request handler for the quote of the day RSS feed."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "application/rss+xml",
        "application/xml",
    )

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests."""
        if head:
            return
        today = datetime.now(timezone.utc).date()
        quotes = (
            await self.get_quote_of_today(),
            *[
                await self.get_quote_by_date(today - timedelta(days=i))
                for i in range(1, QUOTE_COUNT_TO_SHOW_IN_FEED)
            ],
        )
        await self.render(
            "rss/quote_of_the_day.xml",
            quotes=tuple(q for q in quotes if q),
        )


class QuoteOfTheDayAPI(APIRequestHandler, QuoteOfTheDayBaseHandler):
    """Handler for the JSON API that returns the quote of the day."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
        "text/plain",
    )

    async def get(
        self,
        date_str: None | str = None,
        *,
        head: bool = False,  # pylint: disable=unused-argument
    ) -> None:
        """Handle GET requests."""
        quote_data = await (
            self.get_quote_by_date(date_str)
            if date_str
            else self.get_quote_of_today()
        )

        if not quote_data:
            raise HTTPError(404 if date_str else 503)

        wrong_quote: WrongQuote = quote_data.wrong_quote

        if self.content_type == "text/plain":
            return await self.finish(str(wrong_quote))

        if self.request.path.endswith("/full"):
            return await self.finish(quote_data.to_json())

        await self.finish_dict(
            date=quote_data.date.isoformat(),
            url=quote_data.get_quote_url(),
            id=wrong_quote.get_id_as_str(),
            quote=str(wrong_quote.quote),
            author=str(wrong_quote.author),
            rating=wrong_quote.rating,
        )


class QuoteOfTheDayRedirect(QuoteOfTheDayBaseHandler):
    """Redirect to the quote of the day."""

    async def get(
        self,
        date_str: None | str = None,
        *,
        head: bool = False,  # pylint: disable=unused-argument
    ) -> None:
        """Handle GET requests."""
        wrong_quote_data = await (
            self.get_quote_by_date(date_str)
            if date_str
            else self.get_quote_of_today()
        )

        if not wrong_quote_data:
            raise HTTPError(404 if date_str else 503)

        self.redirect(
            self.fix_url(
                wrong_quote_data.get_quote_url(),
            ),
            False,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the troet page."""

from __future__ import annotations

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_html_response,
    assert_valid_redirect,
    fetch,
)
from .test_settings import parse_cookie


async def test_troet_page(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the troet page."""
    assert_valid_html_response(await fetch("/troet"))

    for save in ("sure", "nope"):
        response = assert_valid_html_response(
            await fetch(f"/troet?text=xyz&save={save}")
        )
        assert response.request.url == response.effective_url
        assert b'<textarea name="text">xyz</textarea>' in response.body
        assert "Set-Cookie" not in response.headers

    response = assert_valid_html_response(
        await fetch("/troet?instance=http://example.invalid&save=nope")
    )
    assert response.request.url == response.effective_url
    assert b'value="http://example.invalid"' in response.body
    assert "Set-Cookie" not in response.headers

    response = assert_valid_html_response(
        await fetch("/troet?instance=example.test&save=sure")
    )
    assert response.request.url == response.effective_url
    assert b'value="https://example.test"' in response.body
    assert "Set-Cookie" in response.headers
    cookies = response.headers.get_list("Set-Cookie")
    assert cookies and len(cookies) == 1
    cookie = parse_cookie(cookies[0])
    assert len(cookie) == 1
    morsel = cookie[tuple(cookie)[0]]
    assert morsel["expires"]
    assert morsel["path"] == "/troet"
    assert morsel["samesite"] == "Strict"
    assert morsel.key == "mastodon-instance"
    assert morsel.value == "https://example.test"

    await assert_valid_redirect(
        fetch,
        "/troet?text=xyz",
        "https://example.test/share?text=xyz",
        {307},
        headers={"Cookie": cookies[0]},
    )
    await assert_valid_redirect(
        fetch,
        "/troet?text=xyz%20123",
        "https://example.test/share?text=xyz+123",
        {307},
        headers={"Cookie": cookies[0]},
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Logging stuff used by the website."""

from __future__ import annotations

import asyncio
import logging
import traceback
from asyncio import AbstractEventLoop
from collections.abc import Coroutine, Iterable
from concurrent.futures import Future
from datetime import datetime, tzinfo
from logging import LogRecord
from pathlib import Path
from typing import Never

import orjson as json
from tornado.httpclient import AsyncHTTPClient

from an_website import DIR as AN_WEBSITE_DIR

from .. import CA_BUNDLE_PATH

HOME: str = Path("~/").expanduser().as_posix().rstrip("/")


def minify_filepath(path: str) -> str:
    """Make a filepath smaller."""
    if path.startswith(f"{HOME}/"):
        return "~" + path.removeprefix(HOME)
    return path


def get_minimal_traceback(
    record: LogRecord, prefix: str = "\n\n"
) -> Iterable[str]:
    """Get a minimal traceback from the log record."""
    if not record.exc_info:
        return
    (_, value, tb) = record.exc_info
    if not (value and tb):
        return

    yield prefix
    yield from traceback.format_exception(value, limit=0)

    summary = traceback.extract_tb(tb)
    if isinstance(AN_WEBSITE_DIR, Path):
        start_path = f"{str(AN_WEBSITE_DIR).rstrip('/')}/"

        for i in reversed(range(len(summary))):
            if summary[i].filename.startswith(start_path):
                summary = traceback.StackSummary(summary[i:])
                break

    for frame in summary:
        frame.filename = minify_filepath(frame.filename)

    yield from summary.format()


class AsyncHandler(logging.Handler):
    """A logging handler that can handle log records asynchronously."""

    futures: set[Future[object]]
    loop: AbstractEventLoop

    def __init__(
        self,
        level: int | str = logging.NOTSET,
        *,
        loop: AbstractEventLoop,
    ):
        """Initialize the handler."""
        super().__init__(level=level)
        self.futures = set()
        self.loop = loop

    def callback(self, future: Future[object]) -> None:
        """Remove the reference to the future from the handler."""
        self.acquire()
        try:
            self.futures.discard(future)
        finally:
            self.release()

    def emit(  # type: ignore[override]
        self, record: LogRecord
    ) -> None | Coroutine[None, Never, object]:
        """
        Do whatever it takes to actually log the specified logging record.

        This version is intended to be implemented by subclasses and so
        raises a NotImplementedError.
        """
        raise NotImplementedError(
            "emit must be implemented by AsyncHandler subclasses"
        )

    def handle(  # type: ignore[override]
        self, record: LogRecord
    ) -> bool | LogRecord:
        """Handle incoming log records."""
        rv = self.filter(record)
        if isinstance(rv, LogRecord):
            record = rv
        if rv and not self.loop.is_closed():
            self.acquire()
            try:
                if awaitable := self.emit(record):
                    future: Future[object] = asyncio.run_coroutine_threadsafe(
                        awaitable, self.loop
                    )
                    self.futures.add(future)
                    future.add_done_callback(self.callback)
            finally:
                self.release()
        return rv


class DatetimeFormatter(logging.Formatter):
    """A logging formatter that formats the time using datetime."""

    timezone: None | tzinfo = None

    def formatTime(  # noqa: N802
        self, record: LogRecord, datefmt: None | str = None
    ) -> str:
        """Return the creation time of the LogRecord as formatted text."""
        spam = datetime.fromtimestamp(record.created).astimezone(self.timezone)
        if datefmt:
            return spam.strftime(datefmt)
        return spam.isoformat()


class WebhookFormatter(DatetimeFormatter):
    """A logging formatter optimized for logging to a webhook."""

    escape_message = False
    max_message_length: int | None = None

    def format(self, record: LogRecord) -> str:
        """Format the specified record as text."""
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        if (
            self.max_message_length is not None
            and len(record.message) > self.max_message_length
        ):
            record.message = record.message[: self.max_message_length]
        for line in get_minimal_traceback(record):
            if (
                self.max_message_length is not None
                and len(line) + len(record.message) > self.max_message_length
            ):
                if len("...") + len(record.message) <= self.max_message_length:
                    record.message += "..."
                break
            record.message += line
        if self.escape_message:
            record.message = json.dumps(record.message).decode("UTF-8")[1:-1]
        return self.formatMessage(record)


class WebhookHandler(AsyncHandler):
    """A logging handler that sends logs to a webhook."""

    url: str
    content_type: str

    def __init__(
        self,
        level: int | str = logging.NOTSET,
        *,
        loop: AbstractEventLoop,
        url: str,
        content_type: str,
    ):
        """Initialize the handler."""
        super().__init__(level=level, loop=loop)
        self.url = url
        self.content_type = content_type

    async def emit(self, record: LogRecord) -> None:  # type: ignore[override]
        """Send the request to the webhook."""
        # pylint: disable=invalid-overridden-method
        try:
            message = self.format(record)
            await AsyncHTTPClient().fetch(
                self.url,
                method="POST",
                headers={"Content-Type": self.content_type},
                body=message.strip(),
                ca_certs=CA_BUNDLE_PATH,
            )
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the random text API."""

from __future__ import annotations

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_html_response,
    assert_valid_response,
    fetch,
)

PATH = "/api/zufaelliger-text"


async def test_random_text_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the random text API."""
    assert_valid_response(await fetch(PATH), "text/plain", {200})

    assert_valid_html_response(
        await fetch(PATH + "?words=10", headers={"Accept": "text/html"}),
        codes={200},
        assert_canonical=False,
    )


async def test_random_text_seed(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the random text API."""
    seeded_path = f"{PATH}?seed=xyz"

    assert (await fetch(seeded_path)).body == (await fetch(seeded_path)).body

    assert (await fetch(seeded_path, headers={"Accept": "text/html"})).body == (
        await fetch(seeded_path, headers={"Accept": "text/html"})
    ).body
